from typing import Literal, Union

import numpy as np
import pytest
import torch
from torch import nn

from careamics.config import VAEBasedAlgorithm
from careamics.config.architectures import LVAEModel
from careamics.config.likelihood_model import GaussianLikelihoodConfig
from careamics.config.loss_model import LVAELossConfig
from careamics.models.model_factory import model_factory


# TODO move to conftest as a fixture
def create_LVAE_model(
    input_shape: tuple = (1, 64, 64),
    z_dims: list[int] = (128, 128, 128, 128),
    encoder_conv_strides=(2, 2),
    decoder_conv_strides=(2, 2),
    multiscale_count: int = 0,
    output_channels: int = 1,
    analytical_kl: bool = False,
    predict_logvar: Union[Literal["pixelwise"], None] = None,
) -> nn.Module:
    lvae_model_config = LVAEModel(
        architecture="LVAE",
        input_shape=input_shape,
        z_dims=z_dims,
        encoder_conv_strides=encoder_conv_strides,
        decoder_conv_strides=decoder_conv_strides,
        multiscale_count=multiscale_count,
        output_channels=output_channels,
        predict_logvar=predict_logvar,
        analytical_kl=analytical_kl,
    )

    config = VAEBasedAlgorithm(
        algorithm_type="vae",
        algorithm="musplit",
        loss=LVAELossConfig(loss_type="musplit"),
        model=lvae_model_config,
        gaussian_likelihood=GaussianLikelihoodConfig(
            predict_logvar=predict_logvar, logvar_lowerbound=0.0
        ),
    )
    return model_factory(config.model)


@pytest.mark.skip(reason="Needs to be updated")
@pytest.mark.parametrize(
    "img_size, encoder_conv_strides, decoder_conv_strides",
    [
        ([1, 64, 64], [2, 2], [2, 2]),
        ([1, 128, 128], [2, 2], [2, 2]),
        ([1, 16, 64, 64], [1, 2, 2], [1, 2, 2]),
        ([1, 16, 128, 128], [1, 2, 2], [1, 2, 2]),
        ([1, 16, 64, 64], [1, 2, 2], [2, 2]),
    ],
)
def test_first_bottom_up(
    img_size: int,
    encoder_conv_strides,
    decoder_conv_strides,
    tmp_path,
    create_dummy_noise_model,
) -> None:
    model = create_LVAE_model(
        tmp_path=tmp_path,
        encoder_conv_strides=encoder_conv_strides,
        decoder_conv_strides=decoder_conv_strides,
        create_dummy_noise_model=create_dummy_noise_model,
        input_shape=img_size,
    )
    first_bottom_up = model.first_bottom_up
    inputs = torch.ones((1, *img_size))
    output = first_bottom_up(inputs)
    assert output.shape == (1, model.encoder_n_filters, *img_size[1:])


@pytest.mark.skip(reason="Needs to be updated")
@pytest.mark.parametrize(
    "img_size, z_dims, multiscale_count, encoder_conv_stride, decoder_conv_stride",
    [
        ([1, 64, 64], [128, 128, 128], 0, (2, 2), (2, 2)),
        ([1, 64, 64], [128, 128, 128], 2, (2, 2), (2, 2)),
        ([1, 128, 128], [128, 128, 128], 0, (2, 2), (2, 2)),
        ([1, 128, 128], [128, 128, 128], 2, (2, 2), (2, 2)),
        ([1, 64, 64], [128, 128, 128, 128], 0, (2, 2), (2, 2)),
        ([1, 64, 64], [128, 128, 128, 128], 3, (2, 2), (2, 2)),
        ([1, 128, 128], [128, 128, 128, 128], 0, (2, 2), (2, 2)),
        ([1, 128, 128], [128, 128, 128, 128], 3, (2, 2), (2, 2)),
        ([1, 16, 64, 64], [128, 128, 128], 0, (1, 2, 2), (1, 2, 2)),
        ([1, 16, 64, 64], [128, 128, 128], 2, (1, 2, 2), (1, 2, 2)),
        ([1, 16, 128, 128], [128, 128, 128], 0, (1, 2, 2), (1, 2, 2)),
        ([1, 16, 128, 128], [128, 128, 128], 2, (1, 2, 2), (1, 2, 2)),
        ([1, 16, 64, 64], [128, 128, 128, 128], 0, (1, 2, 2), (1, 2, 2)),
        ([1, 16, 64, 64], [128, 128, 128, 128], 3, (1, 2, 2), (1, 2, 2)),
        ([1, 16, 128, 128], [128, 128, 128, 128], 0, (1, 2, 2), (1, 2, 2)),
        ([1, 16, 128, 128], [128, 128, 128, 128], 3, (1, 2, 2), (1, 2, 2)),
    ],
)
def test_bottom_up_layers(
    img_size: list[int],
    z_dims: list[int],
    multiscale_count,
    encoder_conv_stride,
    decoder_conv_stride,
    tmp_path,
    create_dummy_noise_model,
) -> None:
    model = create_LVAE_model(
        tmp_path=tmp_path,
        input_shape=img_size,
        create_dummy_noise_model=create_dummy_noise_model,
        z_dims=z_dims,
        multiscale_count=multiscale_count,
        encoder_conv_strides=encoder_conv_stride,
        decoder_conv_strides=decoder_conv_stride,
    )
    bottom_up_layers = model.bottom_up_layers
    assert len(bottom_up_layers) == len(z_dims)
    n_filters = model.encoder_n_filters
    downscale_factor = [2 for _ in model.image_size[-2:]]
    if len(model.image_size) == 4:
        downscale_factor.insert(0, 1)
    downscale_factor = (1, 1, *downscale_factor)
    expected_img_size = np.array((1, n_filters, *model.image_size[1:])) // np.array(
        downscale_factor
    )
    input_size = (1, n_filters, *model.image_size[1:])
    inputs = torch.ones(input_size)
    for layer in bottom_up_layers:
        input1, input2 = layer(inputs)
        assert input1.shape == tuple(expected_img_size)
        assert input2.shape == tuple(expected_img_size)
        # this is cuz /= gives ufunc error
        # expected_img_size = expected_img_size / np.array(downscale_factor)
        # TODO why do we need to do this?


@pytest.mark.skip(reason="Needs to be updated")
@pytest.mark.parametrize(
    "z_dims, multiscale_count",
    [
        ([128, 128], 4),
        ([128, 128, 128], 5),
    ],
)
def test_LC_init(
    z_dims: list[int],
    multiscale_count: int,
    tmp_path,
    create_dummy_noise_model,
) -> None:

    with pytest.raises(AssertionError):
        create_LVAE_model(
            tmp_path=tmp_path,
            create_dummy_noise_model=create_dummy_noise_model,
            z_dims=z_dims,
            multiscale_count=multiscale_count,
        )


@pytest.mark.skip(reason="Needs to be updated")
@pytest.mark.parametrize(
    "img_size, z_dims, multiscale_count, encoder_conv_stride, decoder_conv_stride",
    [
        ([1, 64, 64], [128, 128, 128], 0, (2, 2), (2, 2)),
        ([1, 64, 64], [128, 128, 128], 2, (2, 2), (2, 2)),
        ([1, 128, 128], [128, 128, 128], 0, (2, 2), (2, 2)),
        ([1, 128, 128], [128, 128, 128], 2, (2, 2), (2, 2)),
        ([1, 64, 64], [128, 128, 128, 128], 0, (2, 2), (2, 2)),
        ([1, 64, 64], [128, 128, 128, 128], 3, (2, 2), (2, 2)),
        ([1, 128, 128], [128, 128, 128, 128], 0, (2, 2), (2, 2)),
        ([1, 128, 128], [128, 128, 128, 128], 3, (2, 2), (2, 2)),
        ([1, 16, 64, 64], [128, 128, 128], 0, (1, 2, 2), (1, 2, 2)),
        ([1, 16, 64, 64], [128, 128, 128], 2, (1, 2, 2), (1, 2, 2)),
        ([1, 16, 128, 128], [128, 128, 128], 0, (1, 2, 2), (1, 2, 2)),
        ([1, 16, 128, 128], [128, 128, 128], 2, (1, 2, 2), (1, 2, 2)),
        ([1, 16, 64, 64], [128, 128, 128, 128], 0, (1, 2, 2), (1, 2, 2)),
        ([1, 16, 64, 64], [128, 128, 128, 128], 3, (1, 2, 2), (1, 2, 2)),
        ([1, 16, 128, 128], [128, 128, 128, 128], 0, (1, 2, 2), (1, 2, 2)),
        ([1, 16, 128, 128], [128, 128, 128, 128], 3, (1, 2, 2), (1, 2, 2)),
    ],
)
def test_bottom_up_pass(
    img_size: list[int],
    z_dims: list[int],
    multiscale_count: int,
    encoder_conv_stride,
    decoder_conv_stride,
    tmp_path,
    create_dummy_noise_model,
) -> None:

    model = create_LVAE_model(
        input_shape=img_size,
        tmp_path=tmp_path,
        create_dummy_noise_model=create_dummy_noise_model,
        z_dims=z_dims,
        encoder_conv_strides=encoder_conv_stride,
        decoder_conv_strides=decoder_conv_stride,
        multiscale_count=multiscale_count,
    )
    first_bottom_up_layer = model.first_bottom_up
    lowres_first_bottom_up_layers = model.lowres_first_bottom_ups
    bottom_up_layers = model.bottom_up_layers

    assert len(bottom_up_layers) == len(
        z_dims
    ), "Different number of bottom_up_layers and z_dims"
    # Check we have the right number of lowres_net's
    for i in range(multiscale_count - 1):
        assert bottom_up_layers[i].lowres_net is not None, "Missing lowres_net"

    img_size = model.image_size
    n_filters = model.encoder_n_filters
    inputs = torch.ones((1, *img_size))
    outputs = model._bottomup_pass(
        inp=inputs,
        first_bottom_up=first_bottom_up_layer,
        lowres_first_bottom_ups=lowres_first_bottom_up_layers,
        bottom_up_layers=bottom_up_layers,
    )
    exp_img_size = img_size
    for i in range(len(bottom_up_layers)):
        if i + 1 > multiscale_count - 1:
            exp_img_size[1:] = [s // 2 for s in exp_img_size[1:]]
        assert outputs[i].shape == (1, n_filters, *exp_img_size[1:])


@pytest.mark.skip(reason="Needs to be updated")
@pytest.mark.parametrize("img_size", [64, 128])
@pytest.mark.parametrize("multiscale_count", [1, 3, 5])
def test_topmost_top_down_layer(
    img_size: int, multiscale_count: int, tmp_path, create_dummy_noise_model
) -> None:
    model = create_LVAE_model(
        input_shape=img_size,
        tmp_path=tmp_path,
        create_dummy_noise_model=create_dummy_noise_model,
        multiscale_count=multiscale_count,
    )
    topmost_top_down = model.top_down_layers[-1]
    n_filters = model.encoder_n_filters

    downscaling = 2 ** (model.n_layers + 1 - multiscale_count)
    downscaled_size = img_size // downscaling
    bu_value = torch.ones((1, n_filters, downscaled_size, downscaled_size))
    output, data = topmost_top_down(bu_value=bu_value, inference_mode=True)

    retain_sp_dims = (
        topmost_top_down.retain_spatial_dims and downscaled_size == img_size
    )
    exp_out_size = downscaled_size if retain_sp_dims else 2 * downscaled_size
    expected_out_shape = (1, n_filters, exp_out_size, exp_out_size)
    expected_z_shape = (1, model.z_dims[0], downscaled_size, downscaled_size)
    assert output.shape == expected_out_shape
    assert data["z"].shape == expected_z_shape


@pytest.mark.skip(reason="Needs to be updated")
@pytest.mark.parametrize(
    "img_size, z_dims, multiscale_count, encoder_conv_stride, decoder_conv_stride",
    [
        ((1, 64, 64), (64, 64), 1, (2, 2), (2, 2)),
        ((1, 128, 128), (64, 64), 1, (2, 2), (2, 2)),
        ((1, 64, 64), (64, 64, 64, 64), 3, (2, 2), (2, 2)),
        ((1, 128, 128), (64, 64, 64, 64), 3, (2, 2), (2, 2)),
        ((1, 64, 64), (64, 64, 64, 64, 64), 5, (2, 2), (2, 2)),
        ((1, 1, 64, 64), (64, 64), 1, (1, 2, 2), (1, 2, 2)),
        ((1, 1, 64, 64), (64, 64), 1, (1, 2, 2), (2, 2)),
    ],
)
def test_all_top_down_layers(
    img_size: int,
    z_dims: tuple[int],
    multiscale_count: int,
    encoder_conv_stride,
    decoder_conv_stride,
    tmp_path,
    create_dummy_noise_model,
) -> None:
    model = create_LVAE_model(
        tmp_path=tmp_path,
        create_dummy_noise_model=create_dummy_noise_model,
        input_shape=img_size,
        z_dims=z_dims,
        multiscale_count=multiscale_count,
        encoder_conv_strides=encoder_conv_stride,
        decoder_conv_strides=decoder_conv_stride,
    )
    top_down_layers = model.top_down_layers
    n_filters = model.encoder_n_filters
    downscaling = 2 ** (model.n_layers + 1 - multiscale_count)
    downscaled_size = [img_sz // downscaling for img_sz in img_size[-2:]]
    if len(img_size) == 4:
        downscaled_size = downscaled_size.insert(1, img_size[1])
    inputs = skip_input = None
    bu_value = torch.ones((1, img_size[1], *downscaled_size))
    for i in reversed(range(model.n_layers)):
        td_layer = top_down_layers[i]
        output, data = td_layer(
            input_=inputs,
            bu_value=bu_value,
            inference_mode=True,
            skip_connection_input=skip_input,
        )
        inputs = bu_value = skip_input = output

        retain_sp_dims = (
            td_layer.retain_spatial_dims and downscaled_size == img_size[-2:]
        )
        exp_out_size = (
            downscaled_size if retain_sp_dims else [2 * s for s in downscaled_size]
        )
        expected_out_shape = (1, n_filters, *exp_out_size)
        expected_z_shape = (1, model.z_dims[0], *downscaled_size)
        assert (
            output.shape == expected_out_shape
        ), f"Found problem in layer {i+1}, retain={td_layer.retain_spatial_dims},"
        f"dwsc={downscaled_size}"
        assert data["z"].shape == expected_z_shape
        downscaled_size = exp_out_size


@pytest.mark.skip(reason="Needs to be updated")
@pytest.mark.parametrize("img_size", [64, 128])
@pytest.mark.parametrize("multiscale_count", [1, 3, 5])
def test_final_top_down(
    img_size: int, multiscale_count: int, tmp_path, create_dummy_noise_model
) -> None:
    model = create_LVAE_model(
        tmp_path=tmp_path,
        create_dummy_noise_model=create_dummy_noise_model,
        input_shape=img_size,
        multiscale_count=multiscale_count,
    )
    final_top_down = model.final_top_down
    n_filters = model.encoder_n_filters
    final_upsampling = not model.no_initial_downscaling
    inp_size = img_size // 2 if final_upsampling else img_size
    inputs = torch.ones((1, n_filters, inp_size, inp_size))
    output = final_top_down(inputs)
    expected_out_shape = (1, n_filters, img_size, img_size)
    assert output.shape == expected_out_shape


@pytest.mark.skip(reason="Needs to be updated")
@pytest.mark.parametrize("img_size", [64, 128])
@pytest.mark.parametrize("multiscale_count", [1, 3, 5])
def test_top_down_pass(
    img_size: int, multiscale_count: int, tmp_path, create_dummy_noise_model
) -> None:
    model = create_LVAE_model(
        tmp_path=tmp_path,
        create_dummy_noise_model=create_dummy_noise_model,
        input_shape=img_size,
        multiscale_count=multiscale_count,
    )
    top_down_layers = model.top_down_layers
    final_top_down = model.final_top_down
    n_filters = model.encoder_n_filters
    n_layers = model.n_layers

    # Compute the bu_values for all the layers
    bu_values = []
    td_sizes = []
    curr_size = img_size
    for i in range(n_layers):
        if i + 1 > multiscale_count - 1:
            curr_size //= 2
        td_sizes.append(curr_size)
        bu_values.append(torch.ones((1, n_filters, curr_size, curr_size)))

    output, data = model.topdown_pass(
        top_down_layers=top_down_layers,
        final_top_down_layer=final_top_down,
        bu_values=bu_values,
    )

    expected_out_shape = (1, n_filters, img_size, img_size)
    assert output.shape == expected_out_shape
    for i in range(n_layers):
        expected_z_shape = (1, model.z_dims[i], td_sizes[i], td_sizes[i])
        assert data["z"][i].shape == expected_z_shape


@pytest.mark.skip(reason="Needs to be updated")
@pytest.mark.parametrize("img_size", [64, 128])
@pytest.mark.parametrize("multiscale_count", [1, 3, 5])
@pytest.mark.parametrize("analytical_kl", [False, True])
@pytest.mark.parametrize("batch_size", [1, 8])
def test_KL_shape(
    img_size: int,
    multiscale_count: int,
    analytical_kl: bool,
    batch_size: int,
    tmp_path,
    create_dummy_noise_model,
) -> None:
    model = create_LVAE_model(
        tmp_path=tmp_path,
        create_dummy_noise_model=create_dummy_noise_model,
        input_shape=img_size,
        multiscale_count=multiscale_count,
        analytical_kl=analytical_kl,
    )
    top_down_layers = model.top_down_layers
    final_top_down = model.final_top_down
    n_filters = model.encoder_n_filters
    n_layers = model.n_layers

    # Compute the bu_values for all the layers
    bu_values = []
    td_sizes = []
    curr_size = img_size
    for i in range(n_layers):
        if i + 1 > multiscale_count - 1:
            curr_size //= 2
        td_sizes.append(curr_size)
        bu_values.append(torch.ones((batch_size, n_filters, curr_size, curr_size)))

    _, data = model.topdown_pass(
        top_down_layers=top_down_layers,
        final_top_down_layer=final_top_down,
        bu_values=bu_values,
    )

    exp_keys = [
        "kl",  # samplewise
        "kl_restricted",
        "kl_channelwise",
        "kl_spatial",
    ]
    assert all(k in data.keys() for k in exp_keys)
    for i in range(n_layers):
        expected_z_shape = (batch_size, model.z_dims[i], td_sizes[i], td_sizes[i])
        assert data["z"][i].shape == expected_z_shape
        assert data["kl"][i].shape == (batch_size,)
        if model._restricted_kl:
            assert data["kl_restricted"][i].shape == (batch_size,)
        assert data["kl_channelwise"][i].shape == (batch_size, model.z_dims[i])
        assert data["kl_spatial"][i].shape == (batch_size, td_sizes[i], td_sizes[i])


@pytest.mark.skip(reason="Needs to be updated")
@pytest.mark.parametrize("img_size", [64, 128])
@pytest.mark.parametrize("multiscale_count", [1, 3, 5])
@pytest.mark.parametrize("predict_logvar", [None, "pixelwise"])
@pytest.mark.parametrize("output_channels", [1, 2])
def test_output_layer(
    img_size: int,
    multiscale_count: int,
    predict_logvar: Union[Literal["pixelwise"], None],
    output_channels: int,
    tmp_path,
    create_dummy_noise_model,
) -> None:
    model = create_LVAE_model(
        tmp_path=tmp_path,
        create_dummy_noise_model=create_dummy_noise_model,
        input_shape=img_size,
        multiscale_count=multiscale_count,
        predict_logvar=predict_logvar,
        output_channels=output_channels,
    )
    out_layer = model.output_layer
    n_filters = model.encoder_n_filters
    input_ = torch.ones((1, n_filters, img_size, img_size))
    output = out_layer(input_)

    num_out_ch = output_channels * (2 if predict_logvar == "pixelwise" else 1)
    exp_out_shape = (1, num_out_ch, img_size, img_size)
    assert output.shape == exp_out_shape


@pytest.mark.parametrize(
    "img_size, z_dims, multiscale_count, encoder_conv_stride, decoder_conv_stride",
    [
        ([64, 64], [32, 32, 32], 1, (2, 2), (2, 2)),
        ([64, 64], [32, 32, 32], 2, (2, 2), (2, 2)),
        ([64, 64], [32, 32, 32], 3, (2, 2), (2, 2)),
        ([32, 32], [32, 32, 32], 1, (2, 2), (2, 2)),
        ([32, 32], [32, 32, 32], 2, (2, 2), (2, 2)),
        ([32, 32], [32, 32, 32], 3, (2, 2), (2, 2)),
        ([64, 64], [32, 32, 32, 32], 1, (2, 2), (2, 2)),
        ([64, 64], [32, 32, 32, 32], 3, (2, 2), (2, 2)),
        ([64, 64], [32, 32, 32, 32], 4, (2, 2), (2, 2)),
        ([32, 32], [32, 32, 32, 32], 1, (2, 2), (2, 2)),
        ([32, 32], [32, 32, 32, 32], 3, (2, 2), (2, 2)),
        ([16, 64, 64], [32, 32, 32], 1, (1, 2, 2), (1, 2, 2)),
        ([16, 64, 64], [32, 32, 32], 2, (1, 2, 2), (1, 2, 2)),
        ([16, 64, 64], [32, 32, 32], 3, (1, 2, 2), (1, 2, 2)),
        ([16, 32, 32], [32, 32, 32], 1, (1, 2, 2), (1, 2, 2)),
        ([16, 32, 32], [32, 32, 32], 2, (1, 2, 2), (1, 2, 2)),
        ([16, 32, 32], [32, 32, 32, 32], 1, (1, 2, 2), (1, 2, 2)),
        ([16, 64, 64], [32, 32, 32, 32], 3, (1, 2, 2), (1, 2, 2)),
        ([16, 64, 64], [32, 32, 32, 32], 1, (1, 2, 2), (1, 2, 2)),
        ([16, 64, 64], [32, 32, 32, 32], 3, (1, 2, 2), (1, 2, 2)),
        ((15, 64, 64), [32, 32, 32], 1, (1, 2, 2), (2, 2)),
        ((15, 64, 64), [32, 32, 32], 2, (1, 2, 2), (2, 2)),
        ((15, 64, 64), [32, 32, 32], 3, (1, 2, 2), (2, 2)),
    ],
)  # TODO LC input in channels
def test_lvae(
    img_size: list[int],
    z_dims: list[int],
    multiscale_count: int,
    encoder_conv_stride,
    decoder_conv_stride,
    tmp_path,
    create_dummy_noise_model,
) -> None:
    model = create_LVAE_model(
        input_shape=img_size,
        z_dims=z_dims,
        multiscale_count=multiscale_count,
        encoder_conv_strides=encoder_conv_stride,
        decoder_conv_strides=decoder_conv_stride,
    )
    inputs = torch.ones((1, multiscale_count, *img_size))
    output, td_data = model(inputs)
    assert (
        output.shape == (1, 1, *img_size)
        if len(encoder_conv_stride) == len(decoder_conv_stride)
        else (1, 1, *img_size[1:])
    )
    assert td_data is not None
    assert len(td_data["z"]) == len(z_dims)
    assert len(td_data["kl"]) == len(z_dims)
    assert all(kl is not None for kl in td_data["kl"])



"""Pixel manipulation methods supported by CAREamics."""

from careamics.utils import BaseEnum


class SupportedPixelManipulation(str, BaseEnum):
    """Supported Noise2Void pixel manipulations.

    - Uniform: Replace masked pixel value by a (uniformly) randomly selected neighbor
        pixel value.
    - Median: Replace masked pixel value by the mean of the neighborhood.
    """

    UNIFORM = "uniform"
    MEDIAN = "median"



"""
Logging submodule.

The methods are responsible for the in-console logger.
"""

import logging
import sys
import time
from collections.abc import Generator
from pathlib import Path
from typing import Any, Optional, Union

LOGGERS: dict = {}


def get_logger(
    name: str,
    log_level: int = logging.INFO,
    log_path: Optional[Union[str, Path]] = None,
) -> logging.Logger:
    """
    Create a python logger instance with configured handlers.

    Parameters
    ----------
    name : str
        Name of the logger.
    log_level : int, optional
        Log level (info, error etc.), by default logging.INFO.
    log_path : Optional[Union[str, Path]], optional
        Path in which to save the log, by default None.

    Returns
    -------
    logging.Logger
        Logger.
    """
    logger = logging.getLogger(name)
    logger.propagate = False

    if name in LOGGERS:
        return logger

    for logger_name in LOGGERS:
        if name.startswith(logger_name):
            return logger

    logger.propagate = False

    if log_path:
        handlers = [
            logging.StreamHandler(),
            logging.FileHandler(log_path),
        ]
    else:
        handlers = [logging.StreamHandler()]

    formatter = logging.Formatter("%(message)s")

    for handler in handlers:
        handler.setFormatter(formatter)  # type: ignore
        handler.setLevel(log_level)  # type: ignore
        logger.addHandler(handler)  # type: ignore

    logger.setLevel(log_level)
    LOGGERS[name] = True

    logger.propagate = False

    return logger


class ProgressBar:
    """
    Keras style progress bar.

    Adapted from https://github.com/yueyericardo/pkbar.

    Parameters
    ----------
    max_value : Optional[int], optional
        Maximum progress bar value, by default None.
    epoch : Optional[int], optional
        Zero-indexed current epoch, by default None.
    num_epochs : Optional[int], optional
        Total number of epochs, by default None.
    stateful_metrics : Optional[list], optional
        Iterable of string names of metrics that should *not* be averaged over time.
        Metrics in this list will be displayed as-is. All others will be averaged by
        the progress bar before display, by default None.
    always_stateful : bool, optional
            Whether to set all metrics to be stateful, by default False.
    mode : str, optional
        Mode, one of "train", "val", or "predict", by default "train".
    """

    def __init__(
        self,
        max_value: Optional[int] = None,
        epoch: Optional[int] = None,
        num_epochs: Optional[int] = None,
        stateful_metrics: Optional[list] = None,
        always_stateful: bool = False,
        mode: str = "train",
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        max_value : Optional[int], optional
            Maximum progress bar value, by default None.
        epoch : Optional[int], optional
            Zero-indexed current epoch, by default None.
        num_epochs : Optional[int], optional
            Total number of epochs, by default None.
        stateful_metrics : Optional[list], optional
            Iterable of string names of metrics that should *not* be averaged over time.
            Metrics in this list will be displayed as-is. All others will be averaged by
            the progress bar before display, by default None.
        always_stateful : bool, optional
             Whether to set all metrics to be stateful, by default False.
        mode : str, optional
            Mode, one of "train", "val", or "predict", by default "train".
        """
        self.max_value = max_value
        # Width of the progress bar
        self.width = 30
        self.always_stateful = always_stateful

        if (epoch is not None) and (num_epochs is not None):
            print(f"Epoch: {epoch + 1}/{num_epochs}")

        if stateful_metrics:
            self.stateful_metrics = set(stateful_metrics)
        else:
            self.stateful_metrics = set()

        self._dynamic_display = (
            (hasattr(sys.stdout, "isatty") and sys.stdout.isatty())
            or "ipykernel" in sys.modules
            or "posix" in sys.modules
        )
        self._total_width = 0
        self._seen_so_far = 0
        # We use a dict + list to avoid garbage collection
        # issues found in OrderedDict
        self._values: dict[Any, Any] = {}
        self._values_order: list[Any] = []
        self._start = time.time()
        self._last_update = 0.0
        self.spin = self.spinning_cursor() if self.max_value is None else None
        if mode == "train" and self.max_value is None:
            self.message = "Estimating dataset size"
        elif mode == "val":
            self.message = "Validating"
        elif mode == "predict":
            self.message = "Denoising"

    def update(
        self, current_step: int, batch_size: int = 1, values: Optional[list] = None
    ) -> None:
        """
        Update the progress bar.

        Parameters
        ----------
        current_step : int
            Index of the current step.
        batch_size : int, optional
            Batch size, by default 1.
        values : Optional[list], optional
            Updated metrics values, by default None.
        """
        values = values or []
        for k, v in values:
            # if torch tensor, convert it to numpy
            if str(type(v)) == "<class 'torch.Tensor'>":
                v = v.detach().cpu().numpy()

            if k not in self._values_order:
                self._values_order.append(k)
            if k not in self.stateful_metrics and not self.always_stateful:
                if k not in self._values:
                    self._values[k] = [
                        v * (current_step - self._seen_so_far),
                        current_step - self._seen_so_far,
                    ]
                else:
                    self._values[k][0] += v * (current_step - self._seen_so_far)
                    self._values[k][1] += current_step - self._seen_so_far
            else:
                # Stateful metrics output a numeric value. This representation
                # means "take an average from a single value" but keeps the
                # numeric formatting.
                self._values[k] = [v, 1]

        self._seen_so_far = current_step

        now = time.time()
        info = f" - {(now - self._start):.0f}s"

        prev_total_width = self._total_width
        if self._dynamic_display:
            sys.stdout.write("\b" * prev_total_width)
            sys.stdout.write("\r")
        else:
            sys.stdout.write("\n")

        if self.max_value is not None:
            bar = f"{current_step}/{self.max_value} ["
            progress = float(current_step) / self.max_value
            progress_width = int(self.width * progress)
            if progress_width > 0:
                bar += "=" * (progress_width - 1)
                if current_step < self.max_value:
                    bar += ">"
                else:
                    bar += "="
            bar += "." * (self.width - progress_width)
            bar += "]"
        else:
            bar = (
                f"{self.message} {next(self.spin)}, tile "  # type: ignore
                f"No. {current_step * batch_size}"
            )

        self._total_width = len(bar)
        sys.stdout.write(bar)

        if current_step > 0:
            time_per_unit = (now - self._start) / current_step
        else:
            time_per_unit = 0

        if time_per_unit >= 1 or time_per_unit == 0:
            info += f" {time_per_unit:.0f}s/step"
        elif time_per_unit >= 1e-3:
            info += f" {time_per_unit * 1e3:.0f}ms/step"
        else:
            info += f" {time_per_unit * 1e6:.0f}us/step"

        for k in self._values_order:
            info += f" - {k}:"
            if isinstance(self._values[k], list):
                avg = self._values[k][0] / max(1, self._values[k][1])
                if abs(avg) > 1e-3:
                    info += f" {avg:.4f}"
                else:
                    info += f" {avg:.4e}"
            else:
                info += f" {self._values[k]}s"

        self._total_width += len(info)
        if prev_total_width > self._total_width:
            info += " " * (prev_total_width - self._total_width)

        if self.max_value is not None and current_step >= self.max_value:
            info += "\n"

        sys.stdout.write(info)
        sys.stdout.flush()

        self._last_update = now

    def add(self, n: int, values: Optional[list] = None) -> None:
        """
        Update the progress bar by n steps.

        Parameters
        ----------
        n : int
            Number of steps to increase the progress bar with.
        values : Optional[list], optional
            Updated metrics values, by default None.
        """
        self.update(self._seen_so_far + n, 1, values=values)

    def spinning_cursor(self) -> Generator:
        """
        Generate a spinning cursor animation.

        Taken from https://github.com/manrajgrover/py-spinners/tree/master.

        Returns
        -------
        Generator
            Generator of animation frames.
        """
        while True:
            yield from [
                "▓ ----- ▒",
                "▓ ----- ▒",
                "▓ ----- ▒",
                "▓ ->--- ▒",
                "▓ ->--- ▒",
                "▓ ->--- ▒",
                "▓ -->-- ▒",
                "▓ -->-- ▒",
                "▓ -->-- ▒",
                "▓ --->- ▒",
                "▓ --->- ▒",
                "▓ --->- ▒",
                "▓ ----> ▒",
                "▓ ----> ▒",
                "▓ ----> ▒",
                "▒ ----- ░",
                "▒ ----- ░",
                "▒ ----- ░",
                "▒ ->--- ░",
                "▒ ->--- ░",
                "▒ ->--- ░",
                "▒ -->-- ░",
                "▒ -->-- ░",
                "▒ -->-- ░",
                "▒ --->- ░",
                "▒ --->- ░",
                "▒ --->- ░",
                "▒ ----> ░",
                "▒ ----> ░",
                "▒ ----> ░",
            ]



"""Base model for the various CAREamics architectures."""

from typing import Any

from pydantic import BaseModel


class ArchitectureModel(BaseModel):
    """
    Base Pydantic model for all model architectures.

    The `model_dump` method allows removing the `architecture` key from the model.
    """

    architecture: str
    """Name of the architecture."""

    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        """
        Dump the model as a dictionary, ignoring the architecture keyword.

        Parameters
        ----------
        **kwargs : Any
            Additional keyword arguments from Pydantic BaseModel model_dump method.

        Returns
        -------
        {str: Any}
            Model as a dictionary.
        """
        model_dict = super().model_dump(**kwargs)

        # remove the architecture key
        model_dict.pop("architecture")

        return model_dict



"""Script containing the common basic blocks (nn.Module) reused by the LadderVAE."""

from collections.abc import Iterable
from copy import deepcopy
from typing import Callable, Literal, Optional, Union

import numpy as np
import torch
import torch.nn as nn

from .stochastic import NormalStochasticBlock
from .utils import (
    crop_img_tensor,
    pad_img_tensor,
)

ConvType = Union[nn.Conv2d, nn.Conv3d]
NormType = Union[nn.BatchNorm2d, nn.BatchNorm3d]
DropoutType = Union[nn.Dropout2d, nn.Dropout3d]


class ResidualBlock(nn.Module):
    """
    Residual block with 2 convolutional layers.

    Some architectural notes:
        - The number of input, intermediate, and output channels is the same,
        - Padding is always 'same',
        - The 2 convolutional layers have the same groups,
        - No stride allowed,
        - Kernel sizes must be odd.

    The output isgiven by: `out = gate(f(x)) + x`.
    The presence of the gating mechanism is optional, and f(x) has different
    structures depending on the `block_type` argument.
    Specifically, `block_type` is a string specifying the block's structure, with:
        a = activation
        b = batch norm
        c = conv layer
        d = dropout.
    For example, "bacdbacd" defines a block with 2x[batchnorm, activation, conv, dropout].
    """

    default_kernel_size = (3, 3)

    def __init__(
        self,
        channels: int,
        nonlin: Callable,
        conv_strides: tuple[int] = (2, 2),
        kernel: Union[int, Iterable[int], None] = None,
        groups: int = 1,
        batchnorm: bool = True,
        block_type: str = None,
        dropout: float = None,
        gated: bool = None,
        conv2d_bias: bool = True,
    ):
        """
        Constructor.

        Parameters
        ----------
        channels: int
            The number of input and output channels (they are the same).
        nonlin: Callable
            The non-linearity function used in the block (e.g., `nn.ReLU`).
        kernel: Union[int, Iterable[int]], optional
            The kernel size used in the convolutions of the block.
            It can be either a single integer or a pair of integers defining the squared kernel.
            Default is `None`.
        groups: int, optional
            The number of groups to consider in the convolutions. Default is 1.
        batchnorm: bool, optional
            Whether to use batchnorm layers. Default is `True`.
        block_type: str, optional
            A string specifying the block structure, check class docstring for more info.
            Default is `None`.
        dropout: float, optional
            The dropout probability in dropout layers. If `None` dropout is not used.
            Default is `None`.
        gated: bool, optional
            Whether to use gated layer. Default is `None`.
        conv2d_bias: bool, optional
            Whether to use bias term in convolutions. Default is `True`.
        """
        super().__init__()

        # Set kernel size & padding
        if kernel is None:
            kernel = self.default_kernel_size
        elif isinstance(kernel, int):
            kernel = (kernel, kernel)
        elif len(kernel) != 2:
            raise ValueError("kernel has to be None, int, or an iterable of length 2")
        assert all(k % 2 == 1 for k in kernel), "kernel sizes have to be odd"
        kernel = list(kernel)

        # Define modules
        conv_layer: ConvType = getattr(nn, f"Conv{len(conv_strides)}d")
        norm_layer: NormType = getattr(nn, f"BatchNorm{len(conv_strides)}d")
        dropout_layer: DropoutType = getattr(nn, f"Dropout{len(conv_strides)}d")
        # TODO: same comment as in lvae.py, would be more readable to have `conv_dims`

        modules = []
        if block_type == "cabdcabd":
            for i in range(2):
                conv = conv_layer(
                    channels,
                    channels,
                    kernel[i],
                    padding="same",
                    groups=groups,
                    bias=conv2d_bias,
                )
                modules.append(conv)
                modules.append(nonlin)
                if batchnorm:
                    modules.append(norm_layer(channels))
                if dropout is not None:
                    modules.append(dropout_layer(dropout))
        elif block_type == "bacdbac":
            for i in range(2):
                if batchnorm:
                    modules.append(norm_layer(channels))
                modules.append(nonlin)
                conv = conv_layer(
                    channels,
                    channels,
                    kernel[i],
                    padding="same",
                    groups=groups,
                    bias=conv2d_bias,
                )
                modules.append(conv)
                if dropout is not None and i == 0:
                    modules.append(dropout_layer(dropout))
        elif block_type == "bacdbacd":
            for i in range(2):
                if batchnorm:
                    modules.append(norm_layer(channels))
                modules.append(nonlin)
                conv = conv_layer(
                    channels,
                    channels,
                    kernel[i],
                    padding="same",
                    groups=groups,
                    bias=conv2d_bias,
                )
                modules.append(conv)
                modules.append(dropout_layer(dropout))

        else:
            raise ValueError(f"unrecognized block type '{block_type}'")

        self.gated = gated
        if gated:
            modules.append(
                GateLayer(
                    channels=channels,
                    conv_strides=conv_strides,
                    kernel_size=1,
                    nonlin=nonlin,
                )
            )

        self.block = nn.Sequential(*modules)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Parameters
        ----------
        x : torch.Tensor
            input tensor # TODO add shape

        Returns
        -------
        torch.Tensor
            output tensor # TODO add shape
        """
        out = self.block(x)
        assert (
            out.shape == x.shape
        ), f"output shape: {out.shape} != input shape: {x.shape}"
        return out + x


class ResidualGatedBlock(ResidualBlock):
    """Layer class that implements a residual block with a gating mechanism."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, gated=True)


class GateLayer(nn.Module):
    """
    Layer class that implements a gating mechanism.

    Double the number of channels through a convolutional layer, then use
    half the channels as gate for the other half.
    """

    def __init__(
        self,
        channels: int,
        conv_strides: tuple[int] = (2, 2),
        kernel_size: int = 3,
        nonlin: Callable = nn.LeakyReLU(),
    ):
        super().__init__()
        assert kernel_size % 2 == 1
        pad = kernel_size // 2
        conv_layer: ConvType = getattr(nn, f"Conv{len(conv_strides)}d")
        self.conv = conv_layer(channels, 2 * channels, kernel_size, padding=pad)
        self.nonlin = nonlin

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Parameters
        ----------
        x : torch.Tensor
            input # TODO add shape

        Returns
        -------
        torch.Tensor
            output # TODO add shape
        """
        x = self.conv(x)
        x, gate = torch.chunk(x, 2, dim=1)
        x = self.nonlin(x)  # TODO remove this?
        gate = torch.sigmoid(gate)
        return x * gate


class ResBlockWithResampling(nn.Module):
    """
    Residual block with resampling.

    Residual block that takes care of resampling (i.e. downsampling or upsampling) steps (by a factor 2).
    It is structured as follows:
        1. `pre_conv`: a downsampling or upsampling strided convolutional layer in case of resampling, or
            a 1x1 convolutional layer that maps the number of channels of the input to `inner_channels`.
        2. `ResidualBlock`
        3. `post_conv`: a 1x1 convolutional layer that maps the number of channels to `c_out`.

    Some implementation notes:
    - Resampling is performed through a strided convolution layer at the beginning of the block.
    - The strided convolution block has fixed kernel size of 3x3 and 1 layer of padding with zeros.
    - The number of channels is adjusted at the beginning and end of the block through 1x1 convolutional layers.
    - The number of internal channels is by default the same as the number of output channels, but
      min_inner_channels can override the behaviour.
    """

    def __init__(
        self,
        mode: Literal["top-down", "bottom-up"],
        c_in: int,
        c_out: int,
        conv_strides: tuple[int],
        min_inner_channels: Union[int, None] = None,
        nonlin: Callable = nn.LeakyReLU(),
        resample: bool = False,
        res_block_kernel: Optional[Union[int, Iterable[int]]] = None,
        groups: int = 1,
        batchnorm: bool = True,
        res_block_type: Union[str, None] = None,
        dropout: Union[float, None] = None,
        gated: Union[bool, None] = None,
        conv2d_bias: bool = True,
        # lowres_input: bool = False,
    ):
        """
        Constructor.

        Parameters
        ----------
        mode: Literal["top-down", "bottom-up"]
            The type of resampling performed in the initial strided convolution of the block.
            If "bottom-up" downsampling of a factor 2 is done.
            If "top-down" upsampling of a factor 2 is done.
        c_in: int
            The number of input channels.
        c_out: int
            The number of output channels.
        min_inner_channels: int, optional
            The number of channels used in the inner layer of this module.
            Default is `None`, meaning that the number of inner channels is set to `c_out`.
        nonlin: Callable, optional
            The non-linearity function used in the block. Default is `nn.LeakyReLU`.
        resample: bool, optional
            Whether to perform resampling in the first convolutional layer.
            If `False`, the first convolutional layer just maps the input to a tensor with
            `inner_channels` channels through 1x1 convolution. Default is `False`.
        res_block_kernel: Union[int, Iterable[int]], optional
            The kernel size used in the convolutions of the residual block.
            It can be either a single integer or a pair of integers defining the squared kernel.
            Default is `None`.
        groups: int, optional
            The number of groups to consider in the convolutions. Default is 1.
        batchnorm: bool, optional
            Whether to use batchnorm layers. Default is `True`.
        res_block_type: str, optional
            A string specifying the structure of residual block.
            Check `ResidualBlock` doscstring for more information.
            Default is `None`.
        dropout: float, optional
            The dropout probability in dropout layers. If `None` dropout is not used.
            Default is `None`.
        gated: bool, optional
            Whether to use gated layer. Default is `None`.
        conv2d_bias: bool, optional
            Whether to use bias term in convolutions. Default is `True`.
        """
        super().__init__()
        assert mode in ["top-down", "bottom-up"]

        conv_layer: ConvType = getattr(nn, f"Conv{len(conv_strides)}d")
        transp_conv_layer: ConvType = getattr(nn, f"ConvTranspose{len(conv_strides)}d")

        if min_inner_channels is None:
            min_inner_channels = 0
        # inner_channels is the number of channels used in the inner layers
        # of ResBlockWithResampling
        inner_channels = max(c_out, min_inner_channels)

        # Define first conv layer to change num channels and/or up/downsample
        if resample:
            if mode == "bottom-up":  # downsample
                self.pre_conv = conv_layer(
                    in_channels=c_in,
                    out_channels=inner_channels,
                    kernel_size=3,
                    padding=1,
                    stride=conv_strides,
                    groups=groups,
                    bias=conv2d_bias,
                )
            elif mode == "top-down":  # upsample
                self.pre_conv = transp_conv_layer(
                    in_channels=c_in,
                    kernel_size=3,
                    out_channels=inner_channels,
                    padding=1,  # TODO maybe don't hardcode this?
                    stride=conv_strides,
                    groups=groups,
                    output_padding=1 if len(conv_strides) == 2 else (0, 1, 1),
                    bias=conv2d_bias,
                )
        elif c_in != inner_channels:
            self.pre_conv = conv_layer(
                c_in, inner_channels, 1, groups=groups, bias=conv2d_bias
            )
        else:
            self.pre_conv = None

        # Residual block
        self.res = ResidualBlock(
            channels=inner_channels,
            conv_strides=conv_strides,
            nonlin=nonlin,
            kernel=res_block_kernel,
            groups=groups,
            batchnorm=batchnorm,
            dropout=dropout,
            gated=gated,
            block_type=res_block_type,
            conv2d_bias=conv2d_bias,
        )

        # Define last conv layer to get correct num output channels
        if inner_channels != c_out:
            self.post_conv = conv_layer(
                inner_channels, c_out, 1, groups=groups, bias=conv2d_bias
            )
        else:
            self.post_conv = None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Parameters
        ----------
        x : torch.Tensor
            input # TODO add shape

        Returns
        -------
        torch.Tensor
            output # TODO add shape
        """
        if self.pre_conv is not None:
            x = self.pre_conv(x)

        x = self.res(x)

        if self.post_conv is not None:
            x = self.post_conv(x)
        return x


class TopDownDeterministicResBlock(ResBlockWithResampling):
    """Resnet block for top-down deterministic layers."""

    def __init__(self, *args, upsample: bool = False, **kwargs):
        kwargs["resample"] = upsample
        super().__init__("top-down", *args, **kwargs)


class BottomUpDeterministicResBlock(ResBlockWithResampling):
    """Resnet block for bottom-up deterministic layers."""

    def __init__(self, *args, downsample: bool = False, **kwargs):
        kwargs["resample"] = downsample
        super().__init__("bottom-up", *args, **kwargs)


class BottomUpLayer(nn.Module):
    """
    Bottom-up deterministic layer.

    It consists of one or a stack of `BottomUpDeterministicResBlock`'s.
    The outputs are the so-called `bu_values` that are later used in the Decoder to update the
    generative distributions.

    NOTE: When Lateral Contextualization is Enabled (i.e., `enable_multiscale=True`),
    the low-res lateral input is first fed through a BottomUpDeterministicBlock (BUDB)
    (without downsampling), and then merged to the latent tensor produced by the primary flow
    of the `BottomUpLayer` through the `MergeLowRes` layer. It is meaningful to remark that
    the BUDB that takes care of encoding the low-res input can be either shared with the
    primary flow (and in that case it is the "same_size" BUDB (or stack of BUDBs) -> see `self.net`),
    or can be a deep-copy of the primary flow's BUDB.
    This behaviour is controlled by `lowres_separate_branch` parameter.
    """

    def __init__(
        self,
        n_res_blocks: int,
        n_filters: int,
        conv_strides: tuple[int] = (2, 2),
        downsampling_steps: int = 0,
        nonlin: Optional[Callable] = None,
        batchnorm: bool = True,
        dropout: Optional[float] = None,
        res_block_type: Optional[str] = None,
        res_block_kernel: Optional[int] = None,
        gated: Optional[bool] = None,
        enable_multiscale: bool = False,
        multiscale_lowres_size_factor: Optional[int] = None,
        lowres_separate_branch: bool = False,
        multiscale_retain_spatial_dims: bool = False,
        decoder_retain_spatial_dims: bool = False,
        output_expected_shape: Optional[Iterable[int]] = None,
    ):
        """
        Constructor.

        Parameters
        ----------
        n_res_blocks: int
            Number of `BottomUpDeterministicResBlock` modules stacked in this layer.
        n_filters: int
            Number of channels present through out the layers of this block.
        downsampling_steps: int, optional
            Number of downsampling steps that has to be done in this layer (typically 1).
            Default is 0.
        nonlin: Callable, optional
            The non-linearity function used in the block. Default is `None`.
        batchnorm: bool, optional
            Whether to use batchnorm layers. Default is `True`.
        dropout: float, optional
            The dropout probability in dropout layers. If `None` dropout is not used.
            Default is `None`.
        res_block_type: str, optional
            A string specifying the structure of residual block.
            Check `ResidualBlock` doscstring for more information.
            Default is `None`.
        res_block_kernel: Union[int, Iterable[int]], optional
            The kernel size used in the convolutions of the residual block.
            It can be either a single integer or a pair of integers defining the squared kernel.
            Default is `None`.
        gated: bool, optional
            Whether to use gated layer. Default is `None`.
        enable_multiscale: bool, optional
            Whether to enable multiscale (Lateral Contextualization) or not. Default is `False`.
        multiscale_lowres_size_factor: int, optional
            A factor the expresses the relative size of the primary flow tensor with respect to the
            lower-resolution lateral input tensor. Default in `None`.
        lowres_separate_branch: bool, optional
            Whether the residual block(s) encoding the low-res input should be shared (`False`) or
            not (`True`) with the primary flow "same-size" residual block(s). Default is `False`.
        multiscale_retain_spatial_dims: bool, optional
            Whether to pad the latent tensor resulting from the bottom-up layer's primary flow
            to match the size of the low-res input. Default is `False`.
        decoder_retain_spatial_dims: bool, optional
            Whether in the corresponding top-down layer the shape of tensor is retained between
            input and output. Default is `False`.
        output_expected_shape: Iterable[int], optional
            The expected shape of the layer output (only used if `enable_multiscale == True`).
            Default is `None`.
        """
        super().__init__()

        # Define attributes for Lateral Contextualization
        self.enable_multiscale = enable_multiscale
        self.lowres_separate_branch = lowres_separate_branch
        self.multiscale_retain_spatial_dims = multiscale_retain_spatial_dims
        self.multiscale_lowres_size_factor = multiscale_lowres_size_factor
        self.decoder_retain_spatial_dims = decoder_retain_spatial_dims
        self.output_expected_shape = output_expected_shape
        assert self.output_expected_shape is None or self.enable_multiscale is True

        bu_blocks_downsized = []
        bu_blocks_samesize = []
        for _ in range(n_res_blocks):
            do_resample = False
            if downsampling_steps > 0:
                do_resample = True
                downsampling_steps -= 1
            block = BottomUpDeterministicResBlock(
                conv_strides=conv_strides,
                c_in=n_filters,
                c_out=n_filters,
                nonlin=nonlin,
                downsample=do_resample,
                batchnorm=batchnorm,
                dropout=dropout,
                res_block_type=res_block_type,
                res_block_kernel=res_block_kernel,
                gated=gated,
            )
            if do_resample:
                bu_blocks_downsized.append(block)
            else:
                bu_blocks_samesize.append(block)

        self.net_downsized = nn.Sequential(*bu_blocks_downsized)
        self.net = nn.Sequential(*bu_blocks_samesize)

        # Using the same net for the low resolution (and larger sized image)
        self.lowres_net = self.lowres_merge = None
        if self.enable_multiscale:
            self._init_multiscale(
                n_filters=n_filters,
                conv_strides=conv_strides,
                nonlin=nonlin,
                batchnorm=batchnorm,
                dropout=dropout,
                res_block_type=res_block_type,
            )

        # msg = f'[{self.__class__.__name__}] McEnabled:{int(enable_multiscale)} '
        # if enable_multiscale:
        #     msg += f'McParallelBeam:{int(multiscale_retain_spatial_dims)} McFactor{multiscale_lowres_size_factor}'
        # print(msg)

    def _init_multiscale(
        self,
        nonlin: Callable = None,
        n_filters: int = None,
        conv_strides: tuple[int] = (2, 2),
        batchnorm: bool = None,
        dropout: float = None,
        res_block_type: str = None,
    ) -> None:
        """
        Bottom-up layer's method that initializes the LC modules.

        Defines the modules responsible of merging compressed lateral inputs to the
        outputs of the primary flow at different hierarchical levels in the
        multiresolution approach (LC). Specifically, the method initializes `lowres_net`
        , which is a stack of `BottomUpDeterministicBlock`'s (w/out downsampling) that
        takes care of additionally processing the low-res input, and `lowres_merge`,
        which is the module responsible of merging the compressed lateral input to the
        main flow.

        NOTE: The merge modality is set by default to "residual", meaning that the
        merge layer performs concatenation on dim=1, followed by 1x1 convolution and
        a Residual Gated block.

        Parameters
        ----------
        nonlin: Callable, optional
            The non-linearity function used in the block. Default is `None`.
        n_filters: int
            Number of channels present through out the layers of this block.
        batchnorm: bool, optional
            Whether to use batchnorm layers. Default is `True`.
        dropout: float, optional
            The dropout probability in dropout layers. If `None` dropout is not used.
            Default is `None`.
        res_block_type: str, optional
            A string specifying the structure of residual block.
            Check `ResidualBlock` doscstring for more information.
            Default is `None`.
        """
        self.lowres_net = self.net
        if self.lowres_separate_branch:
            self.lowres_net = deepcopy(self.net)

        self.lowres_merge = MergeLowRes(
            channels=n_filters,
            conv_strides=conv_strides,
            merge_type="residual",
            nonlin=nonlin,
            batchnorm=batchnorm,
            dropout=dropout,
            res_block_type=res_block_type,
            multiscale_retain_spatial_dims=self.multiscale_retain_spatial_dims,
            multiscale_lowres_size_factor=self.multiscale_lowres_size_factor,
        )

    def forward(
        self, x: torch.Tensor, lowres_x: Union[torch.Tensor, None] = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Forward pass.

        Parameters
        ----------
        x: torch.Tensor
            The input of the `BottomUpLayer`, i.e., the input image or the output of the
            previous layer.
        lowres_x: torch.Tensor, optional
            The low-res input used for Lateral Contextualization (LC). Default is `None`.

        NOTE: first returned tensor is used as input for the next BU layer, while the second
        tensor is the bu_value passed to the top-down layer.
        """
        # The input is fed through the residual downsampling block(s)
        primary_flow = self.net_downsized(x)
        # The downsampling output is fed through additional residual block(s)
        primary_flow = self.net(primary_flow)

        # If LC is not used, simply return output of primary-flow
        if self.enable_multiscale is False:
            assert lowres_x is None
            return primary_flow, primary_flow

        if lowres_x is not None:
            # First encode the low-res lateral input
            lowres_flow = self.lowres_net(lowres_x)
            # Then pass the result through the MergeLowRes layer
            merged = self.lowres_merge(primary_flow, lowres_flow)
        else:
            merged = primary_flow

        # NOTE: Explanation of possible cases for the conditionals:
        # - if both are `True` -> `merged` has the same spatial dims as the input (`x`) since
        #   spatial dims are retained by padding `primary_flow` in `MergeLowRes`. This is
        #   OK for the corresp TopDown layer, as it also retains spatial dims.
        # - if both are `False` -> `merged`'s spatial dims are equal to `self.net_downsized(x)`,
        #   since no padding is done in `MergeLowRes` and, instead, the lowres input is cropped.
        #   This is OK for the corresp TopDown layer, as it also halves the spatial dims.
        # - if 1st is `False` and 2nd is `True` -> not a concern, it cannot happen
        #   (see lvae.py, line 111, intialization of `multiscale_decoder_retain_spatial_dims`).
        if (
            self.multiscale_retain_spatial_dims is False
            or self.decoder_retain_spatial_dims is True
        ):
            return merged, merged

        # NOTE: if we reach here, it means that `multiscale_retain_spatial_dims` is `True`,
        # but `decoder_retain_spatial_dims` is `False`, meaning that merging LC preserves
        # the spatial dimensions, but at the same time we don't want to retain the spatial
        # dims in the corresponding top-down layer. Therefore, we need to crop the tensor.
        if self.output_expected_shape is not None:
            expected_shape = self.output_expected_shape
        else:
            fac = self.multiscale_lowres_size_factor
            expected_shape = (merged.shape[-2] // fac, merged.shape[-1] // fac)
            assert merged.shape[-2:] != expected_shape

        # Crop the resulting tensor so that it matches with the Decoder
        value_to_use_in_topdown = crop_img_tensor(merged, expected_shape)
        return merged, value_to_use_in_topdown


class MergeLayer(nn.Module):
    """
    Layer class that merges two or more input tensors.

    Merges two or more (B, C, [Z], Y, X) input tensors by concatenating
    them along dim=1 and passes the result through:
    a) a convolutional 1x1 layer (`merge_type == "linear"`), or
    b) a convolutional 1x1 layer and then a gated residual block (`merge_type == "residual"`), or
    c) a convolutional 1x1 layer and then an ungated residual block (`merge_type == "residual_ungated"`).
    """

    def __init__(
        self,
        merge_type: Literal["linear", "residual", "residual_ungated"],
        channels: Union[int, Iterable[int]],
        conv_strides: tuple[int] = (2, 2),
        nonlin: Callable = nn.LeakyReLU(),
        batchnorm: bool = True,
        dropout: Optional[float] = None,
        res_block_type: Optional[str] = None,
        res_block_kernel: Optional[int] = None,
        conv2d_bias: Optional[bool] = True,
    ):
        """
        Constructor.

        Parameters
        ----------
        merge_type: Literal["linear", "residual", "residual_ungated"]
            The type of merge done in the layer. It can be chosen between "linear",
            "residual", and "residual_ungated". Check the class docstring for more
            information about the behaviour of different merge modalities.
        channels: Union[int, Iterable[int]]
            The number of channels used in the convolutional blocks of this layer.
            If it is an `int`:
                - 1st 1x1 Conv2d: in_channels=2*channels, out_channels=channels
                - (Optional) ResBlock: in_channels=channels, out_channels=channels
            If it is an Iterable (must have `len(channels)==3`):
                - 1st 1x1 Conv2d: in_channels=sum(channels[:-1]),
                out_channels=channels[-1]
                - (Optional) ResBlock: in_channels=channels[-1],
                out_channels=channels[-1]
        conv_strides: tuple, optional
            The strides used in the convolutions. Default is `(2, 2)`.
        nonlin: Callable, optional
            The non-linearity function used in the block. Default is `nn.LeakyReLU`.
        batchnorm: bool, optional
            Whether to use batchnorm layers. Default is `True`.
        dropout: float, optional
            The dropout probability in dropout layers. If `None` dropout is not used.
            Default is `None`.
        res_block_type: str, optional
            A string specifying the structure of residual block.
            Check `ResidualBlock` doscstring for more information.
            Default is `None`.
        res_block_kernel: Union[int, Iterable[int]], optional
            The kernel size used in the convolutions of the residual block.
            It can be either a single integer or a pair of integers defining the squared
            kernel.
            Default is `None`.
        conv2d_bias: bool, optional
            Whether to use bias term in convolutions. Default is `True`.
        """
        super().__init__()
        try:
            iter(channels)
        except TypeError:  # it is not iterable
            channels = [channels] * 3
        else:  # it is iterable
            if len(channels) == 1:
                channels = [channels[0]] * 3

        self.conv_layer: ConvType = getattr(nn, f"Conv{len(conv_strides)}d")

        if merge_type == "linear":
            self.layer = self.conv_layer(
                sum(channels[:-1]), channels[-1], 1, bias=conv2d_bias
            )
        elif merge_type == "residual":
            self.layer = nn.Sequential(
                self.conv_layer(
                    sum(channels[:-1]), channels[-1], 1, padding=0, bias=conv2d_bias
                ),
                ResidualGatedBlock(
                    conv_strides=conv_strides,
                    channels=channels[-1],
                    nonlin=nonlin,
                    batchnorm=batchnorm,
                    dropout=dropout,
                    block_type=res_block_type,
                    kernel=res_block_kernel,
                    conv2d_bias=conv2d_bias,
                ),
            )
        elif merge_type == "residual_ungated":
            self.layer = nn.Sequential(
                self.conv_layer(
                    sum(channels[:-1]), channels[-1], 1, padding=0, bias=conv2d_bias
                ),
                ResidualBlock(
                    conv_strides=conv_strides,
                    channels=channels[-1],
                    nonlin=nonlin,
                    batchnorm=batchnorm,
                    dropout=dropout,
                    block_type=res_block_type,
                    kernel=res_block_kernel,
                    conv2d_bias=conv2d_bias,
                ),
            )

    def forward(self, *args) -> torch.Tensor:

        # Concatenate the input tensors along dim=1
        x = torch.cat(args, dim=1)

        # Pass the concatenated tensor through the conv layer
        x = self.layer(x)

        return x


class MergeLowRes(MergeLayer):
    """
    Child class of `MergeLayer`.

    Specifically designed to merge the low-resolution patches
    that are used in Lateral Contextualization approach.
    """

    def __init__(self, *args, **kwargs):
        self.retain_spatial_dims = kwargs.pop("multiscale_retain_spatial_dims")
        self.multiscale_lowres_size_factor = kwargs.pop("multiscale_lowres_size_factor")
        super().__init__(*args, **kwargs)

    def forward(self, latent: torch.Tensor, lowres: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Parameters
        ----------
        latent: torch.Tensor
            The output latent tensor from previous layer in the LVAE hierarchy.
        lowres: torch.Tensor
            The low-res patch image to be merged to increase the context.
        """
        # TODO: treat (X, Y) and Z differently (e.g., line 762)
        if self.retain_spatial_dims:
            # Pad latent tensor to match lowres tensor's shape
            # Output.shape == Lowres.shape (== Input.shape),
            # where Input is the input to the BU layer
            latent = pad_img_tensor(latent, lowres.shape[2:])
        else:
            # Crop lowres tensor to match latent tensor's shape
            lz, ly, lx = lowres.shape[2:]
            z = lz // self.multiscale_lowres_size_factor
            y = ly // self.multiscale_lowres_size_factor
            x = lx // self.multiscale_lowres_size_factor
            z_pad = (lz - z) // 2
            y_pad = (ly - y) // 2
            x_pad = (lx - x) // 2
            lowres = lowres[:, :, z_pad:-z_pad, y_pad:-y_pad, x_pad:-x_pad]

        return super().forward(latent, lowres)


class SkipConnectionMerger(MergeLayer):
    """Specialized `MergeLayer` module, handles skip connections in the model."""

    def __init__(
        self,
        nonlin: Callable,
        channels: Union[int, Iterable[int]],
        batchnorm: bool,
        dropout: float,
        res_block_type: str,
        conv_strides: tuple[int] = (2, 2),
        merge_type: Literal["linear", "residual", "residual_ungated"] = "residual",
        conv2d_bias: bool = True,
        res_block_kernel: Optional[int] = None,
    ):
        """
        Constructor.

        nonlin: Callable, optional
            The non-linearity function used in the block. Default is `nn.LeakyReLU`.
        channels: Union[int, Iterable[int]]
            The number of channels used in the convolutional blocks of this layer.
            If it is an `int`:
                - 1st 1x1 Conv2d: in_channels=2*channels, out_channels=channels
                - (Optional) ResBlock: in_channels=channels, out_channels=channels
            If it is an Iterable (must have `len(channels)==3`):
                - 1st 1x1 Conv2d: in_channels=sum(channels[:-1]), out_channels=channels[-1]
                - (Optional) ResBlock: in_channels=channels[-1], out_channels=channels[-1]
        batchnorm: bool
            Whether to use batchnorm layers.
        dropout: float
            The dropout probability in dropout layers. If `None` dropout is not used.
        res_block_type: str
            A string specifying the structure of residual block.
            Check `ResidualBlock` doscstring for more information.
        conv_strides: tuple, optional
            The strides used in the convolutions. Default is `(2, 2)`.
        merge_type: Literal["linear", "residual", "residual_ungated"]
            The type of merge done in the layer. It can be chosen between "linear", "residual", and "residual_ungated".
            Check the class docstring for more information about the behaviour of different merge modalities.
        conv2d_bias: bool, optional
            Whether to use bias term in convolutions. Default is `True`.
        res_block_kernel: Union[int, Iterable[int]], optional
            The kernel size used in the convolutions of the residual block.
            It can be either a single integer or a pair of integers defining the squared kernel.
            Default is `None`.
        """
        super().__init__(
            conv_strides=conv_strides,
            channels=channels,
            nonlin=nonlin,
            merge_type=merge_type,
            batchnorm=batchnorm,
            dropout=dropout,
            res_block_type=res_block_type,
            res_block_kernel=res_block_kernel,
            conv2d_bias=conv2d_bias,
        )


class TopDownLayer(nn.Module):
    """Top-down inference layer.

    It includes:
        - Stochastic sampling,
        - Computation of KL divergence,
        - A small deterministic ResNet that performs upsampling.

    NOTE 1:
        The algorithm for generative inference approximately works as follows:
            - p_params = output of top-down layer above
            - bu = inferred bottom-up value at this layer
            - q_params = merge(bu, p_params)
            - z = stochastic_layer(q_params)
            - (optional) get and merge skip connection from prev top-down layer
            - top-down deterministic ResNet

    NOTE 2:
        The Top-Down layer can work in two modes: inference and prediction/generative.
        Depending on the particular mode, it follows distinct behaviours:
        - In inference mode, parameters of q(z_i|z_i+1) are obtained from the inference path,
        by merging outcomes of bottom-up and top-down passes. The exception is the top layer,
        in which the parameters of q(z_L|x) are set as the output of the topmost bottom-up layer.
        - On the contrary in predicition/generative mode, parameters of q(z_i|z_i+1) can be obtained
        once again by merging bottom-up and top-down outputs (CONDITIONAL GENERATION), or it is
        possible to directly sample from the prior p(z_i|z_i+1) (UNCONDITIONAL GENERATION).

    NOTE 3:
        When doing unconditional generation, bu_value is not available. Hence the
        merge layer is not used, and z is sampled directly from p_params.

    NOTE 4:
        If this is the top layer, at inference time, the uppermost bottom-up value
        is used directly as q_params, and p_params are defined in this layer
        (while they are usually taken from the previous layer), and can be learned.
    """

    def __init__(
        self,
        z_dim: int,
        n_res_blocks: int,
        n_filters: int,
        conv_strides: tuple[int],
        is_top_layer: bool = False,
        upsampling_steps: Union[int, None] = None,
        nonlin: Union[Callable, None] = None,
        merge_type: Union[
            Literal["linear", "residual", "residual_ungated"], None
        ] = None,
        batchnorm: bool = True,
        dropout: Union[float, None] = None,
        stochastic_skip: bool = False,
        res_block_type: Union[str, None] = None,
        res_block_kernel: Union[int, None] = None,
        groups: int = 1,
        gated: Union[bool, None] = None,
        learn_top_prior: bool = False,
        top_prior_param_shape: Union[Iterable[int], None] = None,
        analytical_kl: bool = False,
        retain_spatial_dims: bool = False,
        vanilla_latent_hw: Union[Iterable[int], None] = None,
        input_image_shape: Union[tuple[int, int], None] = None,
        normalize_latent_factor: float = 1.0,
        conv2d_bias: bool = True,
        stochastic_use_naive_exponential: bool = False,
    ):
        """
        Constructor.

        Parameters
        ----------
        z_dim: int
            The size of the latent space.
        n_res_blocks: int
            The number of TopDownDeterministicResBlock blocks
        n_filters: int
            The number of channels present through out the layers of this block.
        conv_strides: tuple, optional
            The strides used in the convolutions. Default is `(2, 2)`.
        is_top_layer: bool, optional
            Whether the current layer is at the top of the Decoder hierarchy. Default is `False`.
        upsampling_steps: int, optional
            The number of upsampling steps that has to be done in this layer (typically 1).
            Default is `None`.
        nonlin: Callable, optional
            The non-linearity function used in the block (e.g., `nn.ReLU`). Default is `None`.
        merge_type: Literal["linear", "residual", "residual_ungated"], optional
            The type of merge done in the layer. It can be chosen between "linear", "residual",
            and "residual_ungated". Check the `MergeLayer` class docstring for more information
            about the behaviour of different merging modalities. Default is `None`.
        batchnorm: bool, optional
            Whether to use batchnorm layers. Default is `True`.
        dropout: float, optional
            The dropout probability in dropout layers. If `None` dropout is not used.
            Default is `None`.
        stochastic_skip: bool, optional
            Whether to use skip connections between previous top-down layer's output and this layer's stochastic output.
            Stochastic skip connection allows the previous layer's output has a way to directly reach this hierarchical
            level, hence facilitating the gradient flow during backpropagation. Default is `False`.
        res_block_type: str, optional
            A string specifying the structure of residual block.
            Check `ResidualBlock` documentation for more information.
            Default is `None`.
        res_block_kernel: Union[int, Iterable[int]], optional
            The kernel size used in the convolutions of the residual block.
            It can be either a single integer or a pair of integers defining the squared kernel.
            Default is `None`.
        groups: int, optional
            The number of groups to consider in the convolutions. Default is 1.
        gated: bool, optional
            Whether to use gated layer in `ResidualBlock`. Default is `None`.
        learn_top_prior:
            Whether to set the top prior as learnable.
            If this is set to `False`, in the top-most layer the prior will be N(0,1).
            Otherwise, we will still have a normal distribution whose parameters will be learnt.
            Default is `False`.
        top_prior_param_shape: Iterable[int], optional
            The size of the tensor which expresses the mean and the variance
            of the prior for the top most layer. Default is `None`.
        analytical_kl: bool, optional
            If True, KL divergence is calculated according to the analytical formula.
            Otherwise, an MC approximation using sampled latents is calculated.
            Default is `False`.
        retain_spatial_dims: bool, optional
            If `True`, the size of Encoder's latent space is kept to `input_image_shape` within the topdown layer.
            This implies that the oput spatial size equals the input spatial size.
            To achieve this, we centercrop the intermediate representation.
            Default is `False`.
        vanilla_latent_hw: Iterable[int], optional
            The shape of the latent tensor used for prediction (i.e., it influences the computation of restricted KL).
            Default is `None`.
        input_image_shape: Tuple[int, int], optionalut
            The shape of the input image tensor.
            When `retain_spatial_dims` is set to `True`, this is used to ensure that the shape of this layer
            output has the same shape as the input. Default is `None`.
        normalize_latent_factor: float, optional
            A factor used to normalize the latent tensors `q_params`.
            Specifically, normalization is done by dividing the latent tensor by this factor.
            Default is 1.0.
        conv2d_bias: bool, optional
            Whether to use bias term is the convolutional blocks of this layer.
            Default is `True`.
        stochastic_use_naive_exponential: bool, optional
            If `False`, in the NormalStochasticBlock2d exponentials are computed according
            to the alternative definition provided by `StableExponential` class.
            This should improve numerical stability in the training process.
            Default is `False`.
        """
        super().__init__()

        self.is_top_layer = is_top_layer
        self.z_dim = z_dim
        self.stochastic_skip = stochastic_skip
        self.learn_top_prior = learn_top_prior
        self.analytical_kl = analytical_kl
        self.retain_spatial_dims = retain_spatial_dims
        self.input_image_shape = (
            input_image_shape if len(conv_strides) == 3 else input_image_shape[1:]
        )
        self.latent_shape = self.input_image_shape if self.retain_spatial_dims else None
        self.normalize_latent_factor = normalize_latent_factor
        self._vanilla_latent_hw = vanilla_latent_hw  # TODO: check this, it is not used

        # Define top layer prior parameters, possibly learnable
        if is_top_layer:
            self.top_prior_params = nn.Parameter(
                torch.zeros(top_prior_param_shape), requires_grad=learn_top_prior
            )

        # Upsampling steps left to do in this layer
        ups_left = upsampling_steps

        # Define deterministic top-down block, which is a sequence of deterministic
        # residual blocks with (optional) upsampling.
        block_list = []
        for _ in range(n_res_blocks):
            do_resample = False
            if ups_left > 0:
                do_resample = True
                ups_left -= 1
            block_list.append(
                TopDownDeterministicResBlock(
                    c_in=n_filters,
                    c_out=n_filters,
                    conv_strides=conv_strides,
                    nonlin=nonlin,
                    upsample=do_resample,
                    batchnorm=batchnorm,
                    dropout=dropout,
                    res_block_type=res_block_type,
                    res_block_kernel=res_block_kernel,
                    gated=gated,
                    conv2d_bias=conv2d_bias,
                    groups=groups,
                )
            )
        self.deterministic_block = nn.Sequential(*block_list)

        # Define stochastic block with convolutions

        self.stochastic = NormalStochasticBlock(
            c_in=n_filters,
            c_vars=z_dim,
            c_out=n_filters,
            conv_dims=len(conv_strides),
            transform_p_params=(not is_top_layer),
            vanilla_latent_hw=vanilla_latent_hw,
            use_naive_exponential=stochastic_use_naive_exponential,
        )

        if not is_top_layer:
            # Merge layer: it combines bottom-up inference and top-down
            # generative outcomes to give posterior parameters
            self.merge = MergeLayer(
                channels=n_filters,
                conv_strides=conv_strides,
                merge_type=merge_type,
                nonlin=nonlin,
                batchnorm=batchnorm,
                dropout=dropout,
                res_block_type=res_block_type,
                res_block_kernel=res_block_kernel,
                conv2d_bias=conv2d_bias,
            )

            # Skip connection that goes around the stochastic top-down layer
            if stochastic_skip:
                self.skip_connection_merger = SkipConnectionMerger(
                    channels=n_filters,
                    conv_strides=conv_strides,
                    nonlin=nonlin,
                    batchnorm=batchnorm,
                    dropout=dropout,
                    res_block_type=res_block_type,
                    merge_type=merge_type,
                    conv2d_bias=conv2d_bias,
                    res_block_kernel=res_block_kernel,
                )

    def sample_from_q(
        self,
        input_: torch.Tensor,
        bu_value: torch.Tensor,
        var_clip_max: Optional[float] = None,
        mask: torch.Tensor = None,
    ) -> torch.Tensor:
        """
        Method computes the latent inference distribution q(z_i|z_{i+1}).

        Used for sampling a latent tensor from it.

        Parameters
        ----------
        input_: torch.Tensor
            The input tensor to the layer, which is the output of the top-down layer.
        bu_value: torch.Tensor
            The tensor defining the parameters /mu_q and /sigma_q computed during the
            bottom-up deterministic pass at the correspondent hierarchical layer.
        var_clip_max: float, optional
            The maximum value reachable by the log-variance of the latent distribution.
            Values exceeding this threshold are clipped. Default is `None`.
        mask: Union[None, torch.Tensor], optional
            A tensor that is used to mask the sampled latent tensor. Default is `None`.
        """
        if self.is_top_layer:  # In top layer, we don't merge bu_value with p_params
            q_params = bu_value
        else:
            # NOTE: Here the assumption is that the vampprior is only applied on the top layer.
            n_img_prior = None
            p_params = self.get_p_params(input_, n_img_prior)
            q_params = self.merge(bu_value, p_params)

        sample = self.stochastic.sample_from_q(q_params, var_clip_max)

        if mask:
            return sample[mask]

        return sample

    def get_p_params(
        self,
        input_: torch.Tensor,
        n_img_prior: int,
    ) -> torch.Tensor:
        """Return the parameters of the prior distribution p(z_i|z_{i+1}).

        The parameters depend on the hierarchical level of the layer:
        - if it is the topmost level, parameters are the ones of the prior.
        - else, the input from the layer above is the parameters itself.

        Parameters
        ----------
        input_: torch.Tensor
            The input tensor to the layer, which is the output of the top-down layer above.
        n_img_prior: int
            The number of images to be generated from the unconditional prior distribution p(z_L).
        """
        p_params = None

        # If top layer, define p_params as the ones of the prior p(z_L)
        if self.is_top_layer:
            p_params = self.top_prior_params

            # Sample specific number of images by expanding the prior
            if n_img_prior is not None:
                p_params = p_params.expand(n_img_prior, -1, -1, -1)

        # Else the input from the layer above is p_params itself
        else:
            p_params = input_

        return p_params

    def forward(
        self,
        input_: Union[torch.Tensor, None] = None,
        skip_connection_input: Union[torch.Tensor, None] = None,
        inference_mode: bool = False,
        bu_value: Union[torch.Tensor, None] = None,
        n_img_prior: Union[int, None] = None,
        forced_latent: Union[torch.Tensor, None] = None,
        force_constant_output: bool = False,
        mode_pred: bool = False,
        use_uncond_mode: bool = False,
        var_clip_max: Union[float, None] = None,
    ) -> tuple[torch.Tensor, dict[str, torch.Tensor]]:
        """Forward pass.

        Parameters
        ----------
        input_: torch.Tensor, optional
            The input tensor to the layer, which is the output of the top-down layer.
            Default is `None`.
        skip_connection_input: torch.Tensor, optional
            The tensor brought by the skip connection between the current and the
            previous top-down layer.
            Default is `None`.
        inference_mode: bool, optional
            Whether the layer is in inference mode. See NOTE 2 in class description
            for more info.
            Default is `False`.
        bu_value: torch.Tensor, optional
            The tensor defining the parameters /mu_q and /sigma_q computed during the
            bottom-up deterministic pass
            at the correspondent hierarchical layer. Default is `None`.
        n_img_prior: int, optional
            The number of images to be generated from the unconditional prior
            distribution p(z_L).
            Default is `None`.
        forced_latent: torch.Tensor, optional
            A pre-defined latent tensor. If it is not `None`, than it is used as the
            actual latent tensor and,
            hence, sampling does not happen. Default is `None`.
        force_constant_output: bool, optional
            Whether to copy the first sample (and rel. distrib parameters) over the
            whole batch.
            This is used when doing experiment from the prior - q is not used.
            Default is `False`.
        mode_pred: bool, optional
            Whether the model is in prediction mode. Default is `False`.
        use_uncond_mode: bool, optional
            Whether to use the uncoditional distribution p(z) to sample latents in
            prediction mode.
        var_clip_max: float
            The maximum value reachable by the log-variance of the latent distribution.
            Values exceeding this threshold are clipped.
        """
        # Check consistency of arguments
        inputs_none = input_ is None and skip_connection_input is None
        if self.is_top_layer and not inputs_none:
            raise ValueError("In top layer, inputs should be None")

        p_params = self.get_p_params(input_, n_img_prior)

        # Get the parameters for the latent distribution to sample from
        if inference_mode:  # TODO What's this ? reuse Fede's code?
            if self.is_top_layer:
                q_params = bu_value
                if mode_pred is False:
                    assert p_params.shape[2:] == bu_value.shape[2:], (
                        "Spatial dimensions of p_params and bu_value should match. "
                        f"Instead, we got p_params={p_params.shape[2:]} and "
                        f"bu_value={bu_value.shape[2:]}."
                    )
            else:
                if use_uncond_mode:
                    q_params = p_params
                else:
                    assert p_params.shape[2:] == bu_value.shape[2:], (
                        "Spatial dimensions of p_params and bu_value should match. "
                        f"Instead, we got p_params={p_params.shape[2:]} and "
                        f"bu_value={bu_value.shape[2:]}."
                    )
                    q_params = self.merge(bu_value, p_params)
        else:  # generative mode, q is not used, we sample from p(z_i | z_{i+1})
            q_params = None

        # NOTE: Sampling is done either from q(z_i | z_{i+1}, x) or p(z_i | z_{i+1})
        # depending on the mode (hence, in practice, by checking whether q_params is None).

        # Normalization of latent space parameters for stablity.
        # See Very deep VAEs generalize autoregressive models.
        if self.normalize_latent_factor:
            q_params = q_params / self.normalize_latent_factor

        # Sample (and process) a latent tensor in the stochastic layer
        x, data_stoch = self.stochastic(
            p_params=p_params,
            q_params=q_params,
            forced_latent=forced_latent,
            force_constant_output=force_constant_output,
            analytical_kl=self.analytical_kl,
            mode_pred=mode_pred,
            use_uncond_mode=use_uncond_mode,
            var_clip_max=var_clip_max,
        )
        # Merge skip connection from previous layer
        if self.stochastic_skip and not self.is_top_layer:
            x = self.skip_connection_merger(x, skip_connection_input)
        if self.retain_spatial_dims:
            # NOTE: we assume that one topdown layer will have exactly one upscaling layer.

            # NOTE: in case, in the Bottom-Up layer, LC retains spatial dimensions,
            # we have the following (see `MergeLowRes`):
            # - the "primary-flow" tensor is padded to match the low-res patch size
            #   (e.g., from 32x32 to 64x64)
            # - padded tensor is then merged with the low-res patch (concatenation
            #   along dim=1 + convolution)
            # Therefore, we need to do the symmetric operation here, that is to
            # crop `x` for the same amount we padded it in the correspondent BU layer.

            # NOTE: cropping is done to retain the shape of the input in the output.
            # Therefore we need it only in the case `x` is the same shape of the input,
            # because that's the only case in which we need to retain the shape.
            # Here, it must be strictly greater than half the input shape, which is
            # the case if and only if `x.shape == self.latent_shape`.
            rescale = (
                np.array((1, 2, 2)) if len(self.latent_shape) == 3 else np.array((2, 2))
            )  # TODO better way?
            new_latent_shape = tuple(np.array(self.latent_shape) // rescale)
            if x.shape[-1] > new_latent_shape[-1]:
                x = crop_img_tensor(x, new_latent_shape)
        # TODO: `retain_spatial_dims` is the same for all the TD layers.
        # How to handle the case in which we do not have LC for all layers?
        # The answer is in `self.latent_shape`, which is equal to `input_image_shape`
        # (e.g., (64, 64)) if `retain_spatial_dims` is `True`, else it is `None`.
        # Last top-down block (sequence of residual blocks w\ upsampling)
        x = self.deterministic_block(x)
        # Save some metrics that will be used in the loss computation
        keys = [
            "z",
            "kl_samplewise",
            "kl_samplewise_restricted",
            "kl_spatial",
            "kl_channelwise",
            "logprob_q",
            "qvar_max",
        ]
        data = {k: data_stoch.get(k, None) for k in keys}
        data["q_mu"] = None
        data["q_lv"] = None
        if data_stoch["q_params"] is not None:
            q_mu, q_lv = data_stoch["q_params"]
            data["q_mu"] = q_mu
            data["q_lv"] = q_lv
        return x, data



"""Prediction Lightning data modules."""

from pathlib import Path
from typing import Any, Callable, Literal, Optional, Union

import numpy as np
import pytorch_lightning as L
from numpy.typing import NDArray
from torch.utils.data import DataLoader

from careamics.config import InferenceConfig
from careamics.config.support import SupportedData
from careamics.dataset import (
    InMemoryPredDataset,
    InMemoryTiledPredDataset,
    IterablePredDataset,
    IterableTiledPredDataset,
)
from careamics.dataset.dataset_utils import list_files
from careamics.dataset.tiling.collate_tiles import collate_tiles
from careamics.file_io.read import get_read_func
from careamics.utils import get_logger

PredictDatasetType = Union[
    InMemoryPredDataset,
    InMemoryTiledPredDataset,
    IterablePredDataset,
    IterableTiledPredDataset,
]

logger = get_logger(__name__)


class PredictDataModule(L.LightningDataModule):
    """
    CAREamics Lightning prediction data module.

    The data module can be used with Path, str or numpy arrays. The data can be either
    a folder containing images or a single file.

    To read custom data types, you can set `data_type` to `custom` in `data_config`
    and provide a function that returns a numpy array from a path as
    `read_source_func` parameter. The function will receive a Path object and
    an axies string as arguments, the axes being derived from the `data_config`.

    You can also provide a `fnmatch` and `Path.rglob` compatible expression (e.g.
    "*.czi") to filter the files extension using `extension_filter`.

    Parameters
    ----------
    pred_config : InferenceModel
        Pydantic model for CAREamics prediction configuration.
    pred_data : pathlib.Path or str or numpy.ndarray
        Prediction data, can be a path to a folder, a file or a numpy array.
    read_source_func : Callable, optional
        Function to read custom types, by default None.
    extension_filter : str, optional
        Filter to filter file extensions for custom types, by default "".
    dataloader_params : dict, optional
        Dataloader parameters, by default {}.
    """

    def __init__(
        self,
        pred_config: InferenceConfig,
        pred_data: Union[Path, str, NDArray],
        read_source_func: Optional[Callable] = None,
        extension_filter: str = "",
        dataloader_params: Optional[dict] = None,
    ) -> None:
        """
        Constructor.

        The data module can be used with Path, str or numpy arrays. The data can be
        either a folder containing images or a single file.

        To read custom data types, you can set `data_type` to `custom` in `data_config`
        and provide a function that returns a numpy array from a path as
        `read_source_func` parameter. The function will receive a Path object and
        an axies string as arguments, the axes being derived from the `data_config`.

        You can also provide a `fnmatch` and `Path.rglob` compatible expression (e.g.
        "*.czi") to filter the files extension using `extension_filter`.

        Parameters
        ----------
        pred_config : InferenceModel
            Pydantic model for CAREamics prediction configuration.
        pred_data : pathlib.Path or str or numpy.ndarray
            Prediction data, can be a path to a folder, a file or a numpy array.
        read_source_func : Callable, optional
            Function to read custom types, by default None.
        extension_filter : str, optional
            Filter to filter file extensions for custom types, by default "".
        dataloader_params : dict, optional
            Dataloader parameters, by default {}.

        Raises
        ------
        ValueError
            If the data type is `custom` and no `read_source_func` is provided.
        ValueError
            If the data type is `array` and the input is not a numpy array.
        ValueError
            If the data type is `tiff` and the input is neither a Path nor a str.
        """
        if dataloader_params is None:
            dataloader_params = {}
        if dataloader_params is None:
            dataloader_params = {}
        super().__init__()

        # check that a read source function is provided for custom types
        if pred_config.data_type == SupportedData.CUSTOM and read_source_func is None:
            raise ValueError(
                f"Data type {SupportedData.CUSTOM} is not allowed without "
                f"specifying a `read_source_func` and an `extension_filer`."
            )

        # check correct input type
        if (
            isinstance(pred_data, np.ndarray)
            and pred_config.data_type != SupportedData.ARRAY
        ):
            raise ValueError(
                f"Received a numpy array as input, but the data type was set to "
                f"{pred_config.data_type}. Set the data type "
                f"to {SupportedData.ARRAY} to predict on numpy arrays."
            )

        # and that Path or str are passed, if tiff file type specified
        elif (isinstance(pred_data, Path) or isinstance(pred_config, str)) and (
            pred_config.data_type != SupportedData.TIFF
            and pred_config.data_type != SupportedData.CUSTOM
        ):
            raise ValueError(
                f"Received a path as input, but the data type was neither set to "
                f"{SupportedData.TIFF} nor {SupportedData.CUSTOM}. Set the data type "
                f" to {SupportedData.TIFF} or "
                f"{SupportedData.CUSTOM} to predict on files."
            )

        # configuration data
        self.prediction_config = pred_config
        self.data_type = pred_config.data_type
        self.batch_size = pred_config.batch_size
        self.dataloader_params = dataloader_params

        self.pred_data = pred_data
        self.tile_size = pred_config.tile_size
        self.tile_overlap = pred_config.tile_overlap

        # check if it is tiled
        self.tiled = self.tile_size is not None and self.tile_overlap is not None

        # read source function
        if pred_config.data_type == SupportedData.CUSTOM:
            # mypy check
            assert read_source_func is not None

            self.read_source_func: Callable = read_source_func
        elif pred_config.data_type != SupportedData.ARRAY:
            self.read_source_func = get_read_func(pred_config.data_type)

        self.extension_filter = extension_filter

    def prepare_data(self) -> None:
        """Hook used to prepare the data before calling `setup`."""
        # if the data is a Path or a str
        if not isinstance(self.pred_data, np.ndarray):
            self.pred_files = list_files(
                self.pred_data, self.data_type, self.extension_filter
            )

    def setup(self, stage: Optional[str] = None) -> None:
        """
        Hook called at the beginning of predict.

        Parameters
        ----------
        stage : Optional[str], optional
            Stage, by default None.
        """
        # if numpy array
        if self.data_type == SupportedData.ARRAY:
            if self.tiled:
                self.predict_dataset: PredictDatasetType = InMemoryTiledPredDataset(
                    prediction_config=self.prediction_config,
                    inputs=self.pred_data,
                )
            else:
                self.predict_dataset = InMemoryPredDataset(
                    prediction_config=self.prediction_config,
                    inputs=self.pred_data,
                )
        else:
            if self.tiled:
                self.predict_dataset = IterableTiledPredDataset(
                    prediction_config=self.prediction_config,
                    src_files=self.pred_files,
                    read_source_func=self.read_source_func,
                )
            else:
                self.predict_dataset = IterablePredDataset(
                    prediction_config=self.prediction_config,
                    src_files=self.pred_files,
                    read_source_func=self.read_source_func,
                )

    def predict_dataloader(self) -> DataLoader:
        """
        Create a dataloader for prediction.

        Returns
        -------
        DataLoader
            Prediction dataloader.
        """
        return DataLoader(
            self.predict_dataset,
            batch_size=self.batch_size,
            collate_fn=collate_tiles if self.tiled else None,
            **self.dataloader_params,
        )


def create_predict_datamodule(
    pred_data: Union[str, Path, NDArray],
    data_type: Union[Literal["array", "tiff", "custom"], SupportedData],
    axes: str,
    image_means: list[float],
    image_stds: list[float],
    tile_size: Optional[tuple[int, ...]] = None,
    tile_overlap: Optional[tuple[int, ...]] = None,
    batch_size: int = 1,
    tta_transforms: bool = True,
    read_source_func: Optional[Callable] = None,
    extension_filter: str = "",
    dataloader_params: Optional[dict] = None,
) -> PredictDataModule:
    """Create a CAREamics prediction Lightning datamodule.

    This function is used to explicitly pass the parameters usually contained in an
    `inference_model` configuration.

    Since the lightning datamodule has no access to the model, make sure that the
    parameters passed to the datamodule are consistent with the model's requirements
    and are coherent. This can be done by creating a `Configuration` object beforehand
    and passing its parameters to the different Lightning modules.

    The data module can be used with Path, str or numpy arrays. To use array data, set
    `data_type` to `array` and pass a numpy array to `train_data`.

    By default, CAREamics only supports types defined in
    `careamics.config.support.SupportedData`. To read custom data types, you can set
    `data_type` to `custom` and provide a function that returns a numpy array from a
    path. Additionally, pass a `fnmatch` and `Path.rglob` compatible expression
    (e.g. "*.jpeg") to filter the files extension using `extension_filter`.

    In `dataloader_params`, you can pass any parameter accepted by PyTorch
    dataloaders, except for `batch_size`, which is set by the `batch_size`
    parameter.

    Parameters
    ----------
    pred_data : str or pathlib.Path or numpy.ndarray
        Prediction data.
    data_type : {"array", "tiff", "custom"}
        Data type, see `SupportedData` for available options.
    axes : str
        Axes of the data, chosen among SCZYX.
    image_means : list of float
        Mean values for normalization, only used if Normalization is defined.
    image_stds : list of float
        Std values for normalization, only used if Normalization is defined.
    tile_size : tuple of int, optional
        Tile size, 2D or 3D tile size.
    tile_overlap : tuple of int, optional
        Tile overlap, 2D or 3D tile overlap.
    batch_size : int
        Batch size.
    tta_transforms : bool, optional
        Use test time augmentation, by default True.
    read_source_func : Callable, optional
        Function to read the source data, used if `data_type` is `custom`, by
        default None.
    extension_filter : str, optional
        Filter for file extensions, used if `data_type` is `custom`, by default "".
    dataloader_params : dict, optional
        Pytorch dataloader parameters, by default {}.

    Returns
    -------
    PredictDataModule
        CAREamics prediction datamodule.

    Notes
    -----
    If you are using a UNet model and tiling, the tile size must be
    divisible in every dimension by 2**d, where d is the depth of the model. This
    avoids artefacts arising from the broken shift invariance induced by the
    pooling layers of the UNet. If your image has less dimensions, as it may
    happen in the Z dimension, consider padding your image.
    """
    if dataloader_params is None:
        dataloader_params = {}

    prediction_dict: dict[str, Any] = {
        "data_type": data_type,
        "tile_size": tile_size,
        "tile_overlap": tile_overlap,
        "axes": axes,
        "image_means": image_means,
        "image_stds": image_stds,
        "tta_transforms": tta_transforms,
        "batch_size": batch_size,
    }

    # validate configuration
    prediction_config = InferenceConfig(**prediction_dict)

    # sanity check on the dataloader parameters
    if "batch_size" in dataloader_params:
        # remove it
        del dataloader_params["batch_size"]

    return PredictDataModule(
        pred_config=prediction_config,
        pred_data=pred_data,
        read_source_func=read_source_func,
        extension_filter=extension_filter,
        dataloader_params=dataloader_params,
    )



"""
Module for CLI functionality and entrypoint.

Contains the CLI entrypoint, the `run` function; and first level subcommands `train`
and `predict`. The `conf` subcommand is added through the `app.add_typer` function, and
its implementation is contained in the conf.py file.
"""

from pathlib import Path
from typing import Annotated, Optional

import click
import typer

from ..careamist import CAREamist
from . import conf
from .utils import handle_2D_3D_callback

app = typer.Typer(
    help="Run CAREamics algorithms from the command line, including Noise2Void "
    "and its many variants and cousins",
    pretty_exceptions_show_locals=False,
)
app.add_typer(conf.app, name="conf")


@app.command()
def train(  # numpydoc ignore=PR01
    source: Annotated[
        Path,
        typer.Argument(
            help="Path to a configuration file or a trained model.",
            exists=True,
            file_okay=True,
            dir_okay=False,
        ),
    ],
    train_source: Annotated[
        Path,
        typer.Option(
            "--train-source",
            "-ts",
            help="Path to the training data.",
            exists=True,
            file_okay=True,
            dir_okay=True,
        ),
    ],
    train_target: Annotated[
        Optional[Path],
        typer.Option(
            "--train-target",
            "-tt",
            help="Path to train target data.",
            exists=True,
            file_okay=True,
            dir_okay=True,
        ),
    ] = None,
    val_source: Annotated[
        Optional[Path],
        typer.Option(
            "--val-source",
            "-vs",
            help="Path to validation data.",
            exists=True,
            file_okay=True,
            dir_okay=True,
        ),
    ] = None,
    val_target: Annotated[
        Optional[Path],
        typer.Option(
            "--val-target",
            "-vt",
            help="Path to validation target data.",
            exists=True,
            file_okay=True,
            dir_okay=True,
        ),
    ] = None,
    use_in_memory: Annotated[
        bool,
        typer.Option(
            "--use-in-memory/--not-in-memory",
            "-m/-M",
            help="Use in memory dataset if possible.",
        ),
    ] = True,
    val_percentage: Annotated[
        float,
        typer.Option(help="Percentage of files to use for validation."),
    ] = 0.1,
    val_minimum_split: Annotated[
        int,
        typer.Option(help="Minimum number of files to use for validation,"),
    ] = 1,
    work_dir: Annotated[
        Optional[Path],
        typer.Option(
            "--work-dir",
            "-wd",
            help=("Path to working directory in which to save checkpoints and logs"),
            exists=True,
            file_okay=False,
            dir_okay=True,
        ),
    ] = None,
):
    """Train CAREamics models."""
    engine = CAREamist(source=source, work_dir=work_dir)
    engine.train(
        train_source=train_source,
        val_source=val_source,
        train_target=train_target,
        val_target=val_target,
        use_in_memory=use_in_memory,
        val_percentage=val_percentage,
        val_minimum_split=val_minimum_split,
    )


@app.command()
def predict(  # numpydoc ignore=PR01
    model: Annotated[
        Path,
        typer.Argument(
            help="Path to a configuration file or a trained model.",
            exists=True,
            file_okay=True,
            dir_okay=False,
        ),
    ],
    source: Annotated[
        Path,
        typer.Argument(
            help="Path to the training data. Can be a directory or single file.",
            exists=True,
            file_okay=True,
            dir_okay=True,
        ),
    ],
    batch_size: Annotated[int, typer.Option(help="Batch size.")] = 1,
    tile_size: Annotated[
        Optional[click.Tuple],
        typer.Option(
            help=(
                "Size of the tiles to use for prediction, (if the data "
                "is not 3D pass the last value as -1 e.g. --tile_size 64 64 -1)."
            ),
            click_type=click.Tuple([int, int, int]),
            callback=handle_2D_3D_callback,
        ),
    ] = None,
    tile_overlap: Annotated[
        click.Tuple,
        typer.Option(
            help=(
                "Overlap between tiles, (if the data is not 3D pass the last value as "
                "-1 e.g. --tile_overlap 64 64 -1)."
            ),
            click_type=click.Tuple([int, int, int]),
            callback=handle_2D_3D_callback,
        ),
    ] = (48, 48, -1),
    axes: Annotated[
        Optional[str],
        typer.Option(
            help="Axes of the input data. If unused the data is assumed to have the "
            "same axes as the original training data."
        ),
    ] = None,
    data_type: Annotated[
        click.Choice,
        typer.Option(click_type=click.Choice(["tiff"]), help="Type of the input data."),
    ] = "tiff",
    tta_transforms: Annotated[
        bool,
        typer.Option(
            "--tta-transforms/--no-tta-transforms",
            "-t/-T",
            help="Whether to apply test-time augmentation.",
        ),
    ] = False,
    write_type: Annotated[
        click.Choice,
        typer.Option(
            click_type=click.Choice(["tiff"]), help="Type of the output data."
        ),
    ] = "tiff",
    # TODO: could make dataloader_params as json, necessary?
    work_dir: Annotated[
        Optional[Path],
        typer.Option(
            "--work-dir",
            "-wd",
            help=("Path to working directory."),
            exists=True,
            file_okay=False,
            dir_okay=True,
        ),
    ] = None,
    prediction_dir: Annotated[
        Path,
        typer.Option(
            "--prediction-dir",
            "-pd",
            help=(
                "Directory to save predictions to. If not an abosulte path it will be "
                "relative to the set working directory."
            ),
            file_okay=False,
            dir_okay=True,
        ),
    ] = Path("predictions"),
):
    """Create and save predictions from CAREamics models."""
    engine = CAREamist(source=model, work_dir=work_dir)
    engine.predict_to_disk(
        source=source,
        batch_size=batch_size,
        tile_size=tile_size,
        tile_overlap=tile_overlap,
        axes=axes,
        data_type=data_type,
        tta_transforms=tta_transforms,
        write_type=write_type,
        prediction_dir=prediction_dir,
    )


def run():
    """CLI Entry point."""
    app()



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Optional

from numpy.typing import NDArray
import numpy as np
import torch
import torch.nn as nn

if TYPE_CHECKING:
    from careamics.config import GaussianMixtureNMConfig, MultiChannelNMConfig

# TODO this module shouldn't be in lvae folder


def create_histogram(
    bins: int, min_val: float, max_val: float, observation: NDArray, signal: NDArray
) -> NDArray:
    """
    Creates a 2D histogram from 'observation' and 'signal'.

    Parameters
    ----------
    bins : int
        Number of bins in x and y.
    min_val : float
        Lower bound of the lowest bin in x and y.
    max_val : float
        Upper bound of the highest bin in x and y.
    observation : np.ndarray
        3D numpy array (stack of 2D images).
        Observation.shape[0] must be divisible by signal.shape[0].
        Assumes that n subsequent images in observation belong to one image in 'signal'.
    signal : np.ndarray
        3D numpy array (stack of 2D images).

    Returns
    -------
    histogram : np.ndarray
        A 3D array:
        - histogram[0]: Normalized 2D counts.
        - histogram[1]: Lower boundaries of bins along y.
        - histogram[2]: Upper boundaries of bins along y.
    The values for x can be obtained by transposing 'histogram[1]' and 'histogram[2]'.
    """
    histogram = np.zeros((3, bins, bins))

    value_range = [min_val, max_val]

    # Compute mapping factor between observation and signal samples
    obs_to_signal_shape_factor = int(observation.shape[0] / signal.shape[0])

    # Flatten arrays and align signal values
    signal_indices = np.arange(observation.shape[0]) // obs_to_signal_shape_factor
    signal_values = signal[signal_indices].ravel()
    observation_values = observation.ravel()

    count_histogram, signal_edges, _ = np.histogram2d(
        signal_values, observation_values, bins=bins, range=[value_range, value_range]
    )

    # Normalize rows to obtain probabilities
    row_sums = count_histogram.sum(axis=1, keepdims=True)
    count_histogram /= np.clip(row_sums, a_min=1e-20, a_max=None)

    histogram[0] = count_histogram
    histogram[1] = signal_edges[:-1][..., np.newaxis]
    histogram[2] = signal_edges[1:][..., np.newaxis]

    return histogram


def noise_model_factory(
    model_config: Optional[MultiChannelNMConfig],
) -> Optional[MultiChannelNoiseModel]:
    """Noise model factory.

    Parameters
    ----------
    model_config : Optional[MultiChannelNMConfig]
        Noise model configuration, a `MultiChannelNMConfig` config that defines
        noise models for the different output channels.

    Returns
    -------
    Optional[MultiChannelNoiseModel]
        A noise model instance.

    Raises
    ------
    NotImplementedError
        If the chosen noise model `model_type` is not implemented.
        Currently only `GaussianMixtureNoiseModel` is implemented.
    """
    if model_config:
        noise_models = []
        for nm in model_config.noise_models:
            if nm.path:
                if nm.model_type == "GaussianMixtureNoiseModel":
                    noise_models.append(GaussianMixtureNoiseModel(nm))
                else:
                    raise NotImplementedError(
                        f"Model {nm.model_type} is not implemented"
                    )

            else:  # TODO this means signal/obs are provided. Controlled in pydantic model
                # TODO train a new model. Config should always be provided?
                if nm.model_type == "GaussianMixtureNoiseModel":
                    # TODO one model for each channel all make this choise inside the model?
                    # trained_nm = train_gm_noise_model(nm)
                    # noise_models.append(trained_nm)
                    raise NotImplementedError(
                        "GaussianMixtureNoiseModel model training is not implemented."
                    )
                else:
                    raise NotImplementedError(
                        f"Model {nm.model_type} is not implemented"
                    )
        return MultiChannelNoiseModel(noise_models)
    return None


def train_gm_noise_model(
    model_config: GaussianMixtureNMConfig,
    signal: np.ndarray,
    observation: np.ndarray,
) -> GaussianMixtureNoiseModel:
    """Train a Gaussian mixture noise model.

    Parameters
    ----------
    model_config : GaussianMixtureNoiseModel
        _description_

    Returns
    -------
    _description_
    """
    # TODO where to put train params?
    # TODO any training params ? Different channels ?
    noise_model = GaussianMixtureNoiseModel(model_config)
    # TODO revisit config unpacking
    noise_model.fit(signal, observation)
    return noise_model


class MultiChannelNoiseModel(nn.Module):
    def __init__(self, nmodels: list[GaussianMixtureNoiseModel]):
        """Constructor.

        To handle noise models and the relative likelihood computation for multiple
        output channels (e.g., muSplit, denoiseSplit).

        This class:
        - receives as input a variable number of noise models, one for each channel.
        - computes the likelihood of observations given signals for each channel.
        - returns the concatenation of these likelihoods.

        Parameters
        ----------
        nmodels : list[GaussianMixtureNoiseModel]
            List of noise models, one for each output channel.
        """
        super().__init__()
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        for i, nmodel in enumerate(nmodels):  # TODO refactor this !!!
            if nmodel is not None:
                self.add_module(
                    f"nmodel_{i}", nmodel
                )  # TODO: wouldn't be easier to use a list?

        self._nm_cnt = 0
        for nmodel in nmodels:
            if nmodel is not None:
                self._nm_cnt += 1

        print(f"[{self.__class__.__name__}] Nmodels count:{self._nm_cnt}")

    def to_device(self, device: torch.device):
        self.device = device
        self.to(device)
        for ch_idx in range(self._nm_cnt):
            nmodel = getattr(self, f"nmodel_{ch_idx}")
            nmodel.to_device(device)

    def likelihood(self, obs: torch.Tensor, signal: torch.Tensor) -> torch.Tensor:
        """Compute the likelihood of observations given signals for each channel.

        Parameters
        ----------
        obs : torch.Tensor
            Noisy observations, i.e., the target(s). Specifically, the input noisy
            image for HDN, or the noisy unmixed images used for supervision
            for denoiSplit. Shape: (B, C, [Z], Y, X), where C is the number of
            unmixed channels.
        signal : torch.Tensor
            Underlying signals, i.e., the (clean) output of the model. Specifically, the
            denoised image for HDN, or the unmixed images for denoiSplit.
            Shape: (B, C, [Z], Y, X), where C is the number of unmixed channels.
        """
        # Case 1: obs and signal have a single channel (e.g., denoising)
        if obs.shape[1] == 1:
            assert signal.shape[1] == 1
            return self.nmodel_0.likelihood(obs, signal)

        # Case 2: obs and signal have multiple channels (e.g., denoiSplit)
        assert obs.shape[1] == self._nm_cnt, (
            "The number of channels in `obs` must match the number of noise models."
            f" Got instead: obs={obs.shape[1]},  nm={self._nm_cnt}"
        )
        ll_list = []
        for ch_idx in range(obs.shape[1]):
            nmodel = getattr(self, f"nmodel_{ch_idx}")
            ll_list.append(
                nmodel.likelihood(
                    obs[:, ch_idx : ch_idx + 1], signal[:, ch_idx : ch_idx + 1]
                )  # slicing to keep the channel dimension
            )
        return torch.cat(ll_list, dim=1)


class GaussianMixtureNoiseModel(nn.Module):
    """Define a noise model parameterized as a mixture of gaussians.

    If `config.path` is not provided a new object is initialized from scratch.
    Otherwise, a model is loaded from `config.path`.

    Parameters
    ----------
    config : GaussianMixtureNMConfig
        A `pydantic` model that defines the configuration of the GMM noise model.

    Attributes
    ----------
    min_signal : float
        Minimum signal intensity expected in the image.
    max_signal : float
        Maximum signal intensity expected in the image.
    path: Union[str, Path]
        Path to the directory where the trained noise model (*.npz) is saved in the `train` method.
    weight : torch.nn.Parameter
        A [3*n_gaussian, n_coeff] sized array containing the values of the weights
        describing the GMM noise model, with each row corresponding to one
        parameter of each gaussian, namely [mean, standard deviation and weight].
        Specifically, rows are organized as follows:
        - first n_gaussian rows correspond to the means
        - next n_gaussian rows correspond to the weights
        - last n_gaussian rows correspond to the standard deviations
        If `weight=None`, the weight array is initialized using the `min_signal`
        and `max_signal` parameters.
    n_gaussian: int
        Number of gaussians in the mixture.
    n_coeff: int
        Number of coefficients to describe the functional relationship between gaussian
        parameters and the signal. 2 implies a linear relationship, 3 implies a quadratic
        relationship and so on.
    device: device
        GPU device.
    min_sigma: float
        All values of `standard deviation` below this are clamped to this value.
    """

    # TODO training a NM relies on getting a clean data(N2V e.g,)
    def __init__(self, config: GaussianMixtureNMConfig) -> None:
        super().__init__()
        self.device = torch.device("cpu")

        if config.path is not None:
            params = np.load(config.path)
        else:
            params = config.model_dump(exclude_none=True)

        min_sigma = torch.tensor(params["min_sigma"])
        min_signal = torch.tensor(params["min_signal"])
        max_signal = torch.tensor(params["max_signal"])
        self.register_buffer("min_signal", min_signal)
        self.register_buffer("max_signal", max_signal)
        self.register_buffer("min_sigma", min_sigma)
        self.register_buffer("tolerance", torch.tensor([1e-10]))

        if "trained_weight" in params:
            weight = torch.tensor(params["trained_weight"])
        elif "weight" in params and params["weight"] is not None:
            weight = torch.tensor(params["weight"])
        else:
            weight = self._initialize_weights(
                params["n_gaussian"], params["n_coeff"], max_signal, min_signal
            )

        self.n_gaussian = weight.shape[0] // 3
        self.n_coeff = weight.shape[1]

        self.register_parameter("weight", nn.Parameter(weight))
        self._set_model_mode(mode="prediction")

        print(f"[{self.__class__.__name__}] min_sigma: {self.min_sigma}")

    def _initialize_weights(
        self,
        n_gaussian: int,
        n_coeff: int,
        max_signal: torch.Tensor,
        min_signal: torch.Tensor,
    ) -> torch.Tensor:
        """Create random weight initialization."""
        weight = torch.randn(n_gaussian * 3, n_coeff)
        weight[n_gaussian : 2 * n_gaussian, 1] = torch.log(
            max_signal - min_signal
        ).float()
        return weight

    def to_device(self, device: torch.device):
        self.device = device
        self.to(device)

    def _set_model_mode(self, mode: str) -> None:
        """Move parameters to the device and set weights' requires_grad depending on the mode"""
        if mode == "train":
            self.weight.requires_grad = True
        else:
            self.weight.requires_grad = False

    def polynomial_regressor(
        self, weight_params: torch.Tensor, signals: torch.Tensor
    ) -> torch.Tensor:
        """Combines `weight_params` and signal `signals` to regress for the gaussian parameter values.

        Parameters
        ----------
        weight_params : Tensor
            Corresponds to specific rows of the `self.weight`

        signals : Tensor
            Signals

        Returns
        -------
        value : Tensor
            Corresponds to either of mean, standard deviation or weight, evaluated at `signals`
        """
        value = torch.zeros_like(signals)
        for i in range(weight_params.shape[0]):
            value += weight_params[i] * (
                ((signals - self.min_signal) / (self.max_signal - self.min_signal)) ** i
            )
        return value

    def normal_density(
        self, x: torch.Tensor, mean: torch.Tensor, std: torch.Tensor
    ) -> torch.Tensor:
        """
        Evaluates the normal probability density at `x` given the mean `mean` and standard deviation `std`.

        Parameters
        ----------
        x: Tensor
            Observations
        mean: Tensor
            Mean
        std: Tensor
            Standard-deviation

        Returns
        -------
        tmp: Tensor
            Normal probability density of `x` given `mean` and `std`
        """
        tmp = -((x - mean) ** 2)
        tmp = tmp / (2.0 * std * std)
        tmp = torch.exp(tmp)
        tmp = tmp / torch.sqrt((2.0 * np.pi) * std * std)
        return tmp

    def likelihood(
        self, observations: torch.Tensor, signals: torch.Tensor
    ) -> torch.Tensor:
        """
        Evaluates the likelihood of observations given the signals and the corresponding gaussian parameters.

        Parameters
        ----------
        observations : Tensor
            Noisy observations
        signals : Tensor
            Underlying signals

        Returns
        -------
        value: torch.Tensor:
            Likelihood of observations given the signals and the GMM noise model
        """
        gaussian_parameters: list[torch.Tensor] = self.get_gaussian_parameters(signals)
        p = 0
        for gaussian in range(self.n_gaussian):
            p += (
                self.normal_density(
                    observations,
                    gaussian_parameters[gaussian],
                    gaussian_parameters[self.n_gaussian + gaussian],
                )
                * gaussian_parameters[2 * self.n_gaussian + gaussian]
            )
        return p + self.tolerance

    def get_gaussian_parameters(self, signals: torch.Tensor) -> list[torch.Tensor]:
        """
        Returns the noise model for given signals

        Parameters
        ----------
        signals : Tensor
            Underlying signals

        Returns
        -------
        noise_model: list of Tensor
            Contains a list of `mu`, `sigma` and `alpha` for the `signals`
        """
        noise_model = []
        mu = []
        sigma = []
        alpha = []
        kernels = self.weight.shape[0] // 3
        for num in range(kernels):
            mu.append(self.polynomial_regressor(self.weight[num, :], signals))
            expval = torch.exp(self.weight[kernels + num, :])
            sigma_temp = self.polynomial_regressor(expval, signals)
            sigma_temp = torch.clamp(sigma_temp, min=self.min_sigma)
            sigma.append(torch.sqrt(sigma_temp))

            expval = torch.exp(
                self.polynomial_regressor(self.weight[2 * kernels + num, :], signals)
                + self.tolerance
            )
            alpha.append(expval)

        sum_alpha = 0
        for al in range(kernels):
            sum_alpha = alpha[al] + sum_alpha

        # sum of alpha is forced to be 1.
        for ker in range(kernels):
            alpha[ker] = alpha[ker] / sum_alpha

        sum_means = 0
        # sum_means is the alpha weighted average of the means
        for ker in range(kernels):
            sum_means = alpha[ker] * mu[ker] + sum_means

        # subtracting the alpha weighted average of the means from the means
        # ensures that the GMM has the inclination to have the mean=signals.
        # its like a residual conection. I don't understand why we need to learn the mean?
        for ker in range(kernels):
            mu[ker] = mu[ker] - sum_means + signals

        for i in range(kernels):
            noise_model.append(mu[i])
        for j in range(kernels):
            noise_model.append(sigma[j])
        for k in range(kernels):
            noise_model.append(alpha[k])

        return noise_model

    @staticmethod
    def _fast_shuffle(series: torch.Tensor, num: int) -> torch.Tensor:
        """Shuffle the inputs randomly num times"""
        length = series.shape[0]
        for _ in range(num):
            idx = torch.randperm(length)
            series = series[idx, :]
        return series

    def get_signal_observation_pairs(
        self,
        signal: NDArray,
        observation: NDArray,
        lower_clip: float,
        upper_clip: float,
    ) -> torch.Tensor:
        """Returns the Signal-Observation pixel intensities as a two-column array

        Parameters
        ----------
        signal : numpy array
            Clean Signal Data
        observation: numpy array
            Noisy observation Data
        lower_clip: float
            Lower percentile bound for clipping.
        upper_clip: float
            Upper percentile bound for clipping.

        Returns
        -------
        noise_model: list of torch floats
            Contains a list of `mu`, `sigma` and `alpha` for the `signals`
        """
        lb = np.percentile(signal, lower_clip)
        ub = np.percentile(signal, upper_clip)
        stepsize = observation[0].size
        n_observations = observation.shape[0]
        n_signals = signal.shape[0]
        sig_obs_pairs = np.zeros((n_observations * stepsize, 2))

        for i in range(n_observations):
            j = i // (n_observations // n_signals)
            sig_obs_pairs[stepsize * i : stepsize * (i + 1), 0] = signal[j].ravel()
            sig_obs_pairs[stepsize * i : stepsize * (i + 1), 1] = observation[i].ravel()
        sig_obs_pairs = sig_obs_pairs[
            (sig_obs_pairs[:, 0] > lb) & (sig_obs_pairs[:, 0] < ub)
        ]
        sig_obs_pairs = sig_obs_pairs.astype(np.float32)
        sig_obs_pairs = torch.from_numpy(sig_obs_pairs)
        return self._fast_shuffle(sig_obs_pairs, 2)

    def fit(
        self,
        signal: NDArray,
        observation: NDArray,
        learning_rate: float = 1e-1,
        batch_size: int = 250000,
        n_epochs: int = 2000,
        lower_clip: float = 0.0,
        upper_clip: float = 100.0,
    ) -> list[float]:
        """Training to learn the noise model from signal - observation pairs.

        Parameters
        ----------
        signal: numpy array
            Clean Signal Data
        observation: numpy array
            Noisy Observation Data
        learning_rate: float
            Learning rate. Default = 1e-1.
        batch_size: int
            Nini-batch size. Default = 250000.
        n_epochs: int
            Number of epochs. Default = 2000.
        lower_clip : int
            Lower percentile for clipping. Default is 0.
        upper_clip : int
            Upper percentile for clipping. Default is 100.
        """
        self._set_model_mode(mode="train")
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.to_device(device)
        optimizer = torch.optim.Adam([self.weight], lr=learning_rate)

        sig_obs_pairs = self.get_signal_observation_pairs(
            signal, observation, lower_clip, upper_clip
        )

        train_losses = []
        counter = 0
        for t in range(n_epochs):
            if (counter + 1) * batch_size >= sig_obs_pairs.shape[0]:
                counter = 0
                sig_obs_pairs = self._fast_shuffle(sig_obs_pairs, 1)

            batch_vectors = sig_obs_pairs[
                counter * batch_size : (counter + 1) * batch_size, :
            ]
            observations = batch_vectors[:, 1].to(self.device)
            signals = batch_vectors[:, 0].to(self.device)

            p = self.likelihood(observations, signals)

            joint_loss = torch.mean(-torch.log(p))
            train_losses.append(joint_loss.item())

            if self.weight.isnan().any() or self.weight.isinf().any():
                print(
                    "NaN or Inf detected in the weights. Aborting training at epoch: ",
                    t,
                )
                break

            if t % 100 == 0:
                last_losses = train_losses[-100:]
                print(t, np.mean(last_losses))

            optimizer.zero_grad()
            joint_loss.backward()
            optimizer.step()
            counter += 1

        self._set_model_mode(mode="prediction")
        self.to_device(torch.device("cpu"))
        print("===================\n")
        return train_losses

    def sample_observation_from_signal(self, signal: NDArray) -> NDArray:
        """
        Sample an instance of observation based on an input signal using a
        learned Gaussian Mixture Model. For each pixel in the input signal,
        samples a corresponding noisy pixel.

        Parameters
        ----------
        signal: numpy array
            Clean 2D signal data.

        Returns
        -------
        observation: numpy array
            An instance of noisy observation data based on the input signal.
        """
        assert len(signal.shape) == 2, "Only 2D inputs are supported."

        signal_tensor = torch.from_numpy(signal).to(torch.float32)
        height, width = signal_tensor.shape

        with torch.no_grad():
            # Get gaussian parameters for each pixel
            gaussian_params = self.get_gaussian_parameters(signal_tensor)
            means = np.array(gaussian_params[: self.n_gaussian])
            stds = np.array(gaussian_params[self.n_gaussian : self.n_gaussian * 2])
            alphas = np.array(gaussian_params[self.n_gaussian * 2 :])

            if self.n_gaussian == 1:
                # Single gaussian case
                observation = np.random.normal(
                    loc=means[0], scale=stds[0], size=(height, width)
                )
            else:
                # Multiple gaussians: sample component for each pixel
                uniform = np.random.rand(1, height, width)
                # Compute cumulative probabilities for component selection
                cumulative_alphas = np.cumsum(
                    alphas, axis=0
                )  # Shape: (n_gaussian, height, width)
                selected_component = np.argmax(
                    uniform < cumulative_alphas, axis=0, keepdims=True
                )

                # For every pixel, choose the corresponding gaussian
                # and get the learned mu and sigma
                selected_mus = np.take_along_axis(means, selected_component, axis=0)
                selected_stds = np.take_along_axis(stds, selected_component, axis=0)
                selected_mus = selected_mus.squeeze(0)
                selected_stds = selected_stds.squeeze(0)

                # Sample from the normal distribution with learned mu and sigma
                observation = np.random.normal(
                    selected_mus, selected_stds, size=(height, width)
                )
        return observation

    def save(self, path: str, name: str) -> None:
        """Save the trained parameters on the noise model.

        Parameters
        ----------
        path : str
            Path to save the trained parameters.
        name : str
            File name to save the trained parameters.
        """
        os.makedirs(path, exist_ok=True)
        np.savez(
            os.path.join(path, name),
            trained_weight=self.weight.numpy(),
            min_signal=self.min_signal.numpy(),
            max_signal=self.max_signal.numpy(),
            min_sigma=self.min_sigma,
        )
        print("The trained parameters (" + name + ") is saved at location: " + path)



"""N2V manipulation functions for PyTorch."""

from typing import Optional

import torch

from .struct_mask_parameters import StructMaskParameters


def _apply_struct_mask_torch(
    patch: torch.Tensor,
    coords: torch.Tensor,
    struct_params: StructMaskParameters,
    rng: Optional[torch.Generator] = None,
) -> torch.Tensor:
    """Apply structN2V masks to patch.

    Each point in `coords` corresponds to the center of a mask. Masks are parameterized
    by `struct_params`, and pixels in the mask (with respect to `coords`) are replaced
    by a random value.

    Note that the structN2V mask is applied in 2D at the coordinates given by `coords`.

    Parameters
    ----------
    patch : torch.Tensor
        Patch to be manipulated, (batch, y, x) or (batch, z, y, x).
    coords : torch.Tensor
        Coordinates of the ROI (subpatch) centers.
    struct_params : StructMaskParameters
        Parameters for the structN2V mask (axis and span).
    rng : torch.Generator, optional
        Random number generator.

    Returns
    -------
    torch.Tensor
        Patch with the structN2V mask applied.
    """
    if rng is None:
        rng = torch.Generator(device=patch.device)

    # Relative axis
    moving_axis = -1 - struct_params.axis

    # Create a mask array
    mask_shape = [1] * len(patch.shape)
    mask_shape[moving_axis] = struct_params.span
    mask = torch.ones(mask_shape, device=patch.device)

    center = torch.tensor(mask.shape, device=patch.device) // 2

    # Mark the center
    mask[tuple(center)] = 0

    # Displacements from center
    displacements = torch.stack(torch.where(mask == 1)) - center.unsqueeze(1)

    # Combine all coords (ndim, npts) with all displacements (ncoords, ndim)
    mix = displacements.T.unsqueeze(-1) + coords.T.unsqueeze(0)
    mix = mix.permute([1, 0, 2]).reshape([mask.ndim, -1]).T

    # Filter out invalid indices
    valid_indices = (mix[:, moving_axis] >= 0) & (
        mix[:, moving_axis] < patch.shape[moving_axis]
    )
    mix = mix[valid_indices]

    # Replace neighboring pixels with random values from a uniform distribution
    random_values = torch.empty(len(mix), device=patch.device).uniform_(
        patch.min().item(), patch.max().item(), generator=rng
    )
    patch[tuple(mix.T.tolist())] = random_values

    return patch


def _odd_jitter_func_torch(step: float, rng: torch.Generator) -> torch.Tensor:
    """
    Randomly sample a jitter to be applied to the masking grid.

    This is done to account for cases where the step size is not an integer.

    Parameters
    ----------
    step : float
        Step size of the grid, output of np.linspace.
    rng : torch.Generator
        Random number generator.

    Returns
    -------
    torch.Tensor
        Array of random jitter to be added to the grid.
    """
    step_floor = torch.floor(torch.tensor(step))
    odd_jitter = (
        step_floor
        if step_floor == step
        else torch.randint(high=2, size=(1,), generator=rng)
    )
    return step_floor if odd_jitter == 0 else step_floor + 1


def _get_stratified_coords_torch(
    mask_pixel_perc: float,
    shape: tuple[int, ...],
    rng: Optional[torch.Generator] = None,
) -> torch.Tensor:
    """
    Generate coordinates of the pixels to mask.

    # TODO add more details
    Randomly selects the coordinates of the pixels to mask in a stratified way, i.e.
    the distance between masked pixels is approximately the same.

    Parameters
    ----------
    mask_pixel_perc : float
        Actual (quasi) percentage of masked pixels across the whole image. Used in
        calculating the distance between masked pixels across each axis.
    shape : tuple[int, ...]
        Shape of the input patch.
    rng : torch.Generator or None
        Random number generator.

    Returns
    -------
    np.ndarray
        Array of coordinates of the masked pixels.
    """
    if rng is None:
        rng = torch.Generator()

    # Calculate the maximum distance between masked pixels. Inversely proportional to
    # the percentage of masked pixels.
    mask_pixel_distance = round((100 / mask_pixel_perc) ** (1 / len(shape)))

    pixel_coords = []
    steps = []

    # loop over dimensions
    for axis_size in shape:
        # number of pixels to mask along the axis
        num_pixels = int(torch.ceil(torch.tensor(axis_size / mask_pixel_distance)))

        # create 1D grid of coordinates for the axis
        axis_pixel_coords = torch.linspace(
            0,
            axis_size - (axis_size // num_pixels),
            num_pixels,
            dtype=torch.int32,
        )

        # calculate the step size between coordinates
        step = (
            axis_pixel_coords[1] - axis_pixel_coords[0]
            if len(axis_pixel_coords) > 1
            else axis_size
        )

        pixel_coords.append(axis_pixel_coords)
        steps.append(step)

    # create a 2D meshgrid of coordinates
    coordinate_grid_list = torch.meshgrid(*pixel_coords, indexing="ij")
    coordinate_grid = torch.stack(
        [g.flatten() for g in coordinate_grid_list], dim=-1
    ).to(rng.device)

    # add a random jitter increment so that the coordinates do not lie on the grid
    random_increment = torch.randint(
        high=int(_odd_jitter_func_torch(float(max(steps)), rng)),
        size=torch.tensor(coordinate_grid.shape).to(rng.device).tolist(),
        generator=rng,
        device=rng.device,
    )
    coordinate_grid += random_increment

    # make sure no coordinate lie outside the range
    return torch.clamp(
        coordinate_grid,
        torch.zeros_like(torch.tensor(shape)).to(device=rng.device),
        torch.tensor([v - 1 for v in shape]).to(device=rng.device),
    )


def uniform_manipulate_torch(
    patch: torch.Tensor,
    mask_pixel_percentage: float,
    subpatch_size: int = 11,
    remove_center: bool = True,
    struct_params: Optional[StructMaskParameters] = None,
    rng: Optional[torch.Generator] = None,
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Manipulate pixels by replacing them with a neighbor values.

    # TODO add more details, especially about batch

    Manipulated pixels are selected unformly selected in a subpatch, away from a grid
    with an approximate uniform probability to be selected across the whole patch.
    If `struct_params` is not None, an additional structN2V mask is applied to the
    data, replacing the pixels in the mask with random values (excluding the pixel
    already manipulated).

    Parameters
    ----------
    patch : torch.Tensor
        Image patch, 2D or 3D, shape (y, x) or (z, y, x). # TODO batch and channel.
    mask_pixel_percentage : float
        Approximate percentage of pixels to be masked.
    subpatch_size : int
        Size of the subpatch the new pixel value is sampled from, by default 11.
    remove_center : bool
        Whether to remove the center pixel from the subpatch, by default False.
    struct_params : StructMaskParameters or None
        Parameters for the structN2V mask (axis and span).
    rng : torch.default_generator or None
        Random number generator.

    Returns
    -------
    tuple[torch.Tensor, torch.Tensor]
        tuple containing the manipulated patch and the corresponding mask.
    """
    if rng is None:
        rng = torch.Generator(device=patch.device)
        # TODO do we need seed ?

    # create a copy of the patch
    transformed_patch = patch.clone()

    # get the coordinates of the pixels to be masked
    subpatch_centers = _get_stratified_coords_torch(
        mask_pixel_percentage, patch.shape, rng
    )
    subpatch_centers = subpatch_centers.to(device=patch.device)

    # TODO refactor with non negative indices?
    # arrange the list of indices to represent the ROI around the pixel to be masked
    roi_span_full = torch.arange(
        -(subpatch_size // 2),
        subpatch_size // 2 + 1,
        dtype=torch.int32,
        device=patch.device,
    )

    # remove the center pixel from the ROI
    roi_span = roi_span_full[roi_span_full != 0] if remove_center else roi_span_full

    # create a random increment to select the replacement value
    # this increment is added to the center coordinates
    random_increment = roi_span[
        torch.randint(
            low=min(roi_span),
            high=max(roi_span) + 1,  # TODO check this, it may exclude one value
            size=subpatch_centers.shape,
            generator=rng,
            device=patch.device,
        )
    ]

    # compute the replacement pixel coordinates
    replacement_coords = torch.clamp(
        subpatch_centers + random_increment,
        torch.zeros_like(torch.tensor(patch.shape)).to(device=patch.device),
        torch.tensor([v - 1 for v in patch.shape]).to(device=patch.device),
    )

    # replace the pixels in the patch
    # tuples and transpose are needed for proper indexing
    replacement_pixels = patch[tuple(replacement_coords.T)]
    transformed_patch[tuple(subpatch_centers.T)] = replacement_pixels

    # create a mask representing the masked pixels
    mask = (transformed_patch != patch).to(dtype=torch.uint8)

    # apply structN2V mask if needed
    if struct_params is not None:
        transformed_patch = _apply_struct_mask_torch(
            transformed_patch, subpatch_centers, struct_params, rng
        )

    return transformed_patch, mask


def median_manipulate_torch(
    batch: torch.Tensor,
    mask_pixel_percentage: float,
    subpatch_size: int = 11,
    struct_params: Optional[StructMaskParameters] = None,
    rng: Optional[torch.Generator] = None,
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Manipulate pixels by replacing them with the median of their surrounding subpatch.

    N2V2 version, manipulated pixels are selected randomly away from a grid with an
    approximate uniform probability to be selected across the whole patch.

    If `struct_params` is not None, an additional structN2V mask is applied to the data,
    replacing the pixels in the mask with random values (excluding the pixel already
    manipulated).

    Parameters
    ----------
    batch : torch.Tensor
        Image patch, 2D or 3D, shape (y, x) or (z, y, x).
    mask_pixel_percentage : float
        Approximate percentage of pixels to be masked.
    subpatch_size : int
        Size of the subpatch the new pixel value is sampled from, by default 11.
    struct_params : StructMaskParameters or None, optional
        Parameters for the structN2V mask (axis and span).
    rng : torch.default_generator or None, optional
        Random number generato, by default None.

    Returns
    -------
    tuple[torch.Tensor, torch.Tensor, torch.Tensor]
           tuple containing the manipulated patch, the original patch and the mask.
    """
    # get the coordinates of the future ROI centers
    subpatch_center_coordinates = _get_stratified_coords_torch(
        mask_pixel_percentage, batch.shape, rng
    ).to(
        device=batch.device
    )  # (num_coordinates, batch + num_spatial_dims)

    # Calculate the padding value for the input tensor
    pad_value = subpatch_size // 2

    # Generate all offsets for the ROIs. Iteration starting from 1 to skip the batch
    offsets = torch.meshgrid(
        [
            torch.arange(-pad_value, pad_value + 1, device=batch.device)
            for _ in range(1, subpatch_center_coordinates.shape[1])
        ],
        indexing="ij",
    )
    offsets = torch.stack(
        [axis_offset.flatten() for axis_offset in offsets], dim=1
    )  # (subpatch_size**2, num_spatial_dims)

    # Create the list to assemble coordinates of the ROIs centers for each axis
    coords_axes = []
    # Create the list to assemble the span of coordinates defining the ROIs for each
    # axis
    coords_expands = []
    for d in range(subpatch_center_coordinates.shape[1]):
        coords_axes.append(subpatch_center_coordinates[:, d])
        if d == 0:
            # For batch dimension coordinates are not expanded (no offsets)
            coords_expands.append(
                subpatch_center_coordinates[:, d]
                .unsqueeze(1)
                .expand(-1, subpatch_size ** offsets.shape[1])
            )  # (num_coordinates, subpatch_size**num_spacial_dims)
        else:
            # For spatial dimensions, coordinates are expanded with offsets, creating
            # spans
            coords_expands.append(
                (
                    subpatch_center_coordinates[:, d].unsqueeze(1) + offsets[:, d - 1]
                ).clamp(0, batch.shape[d] - 1)
            )  # (num_coordinates, subpatch_size**num_spacial_dims)

    # create array of rois by indexing the batch with gathered coordinates
    rois = batch[
        tuple(coords_expands)
    ]  # (num_coordinates, subpatch_size**num_spacial_dims)

    if struct_params is not None:
        # Create the structN2V mask
        h, w = torch.meshgrid(
            torch.arange(subpatch_size), torch.arange(subpatch_size), indexing="ij"
        )
        center_idx = subpatch_size // 2
        halfspan = (struct_params.span - 1) // 2

        # Determine the axis along which to apply the mask
        if struct_params.axis == 0:
            center_axis = h
            span_axis = w
        else:
            center_axis = w
            span_axis = h

        # Create the mask
        struct_mask = (
            ~(
                (center_axis == center_idx)
                & (span_axis >= center_idx - halfspan)
                & (span_axis <= center_idx + halfspan)
            )
        ).flatten()
        rois_filtered = rois[:, struct_mask]
    else:
        # Remove the center pixel value from the rois
        center_idx = (subpatch_size ** offsets.shape[1]) // 2
        rois_filtered = torch.cat(
            [rois[:, :center_idx], rois[:, center_idx + 1 :]], dim=1
        )

    # compute the medians.
    medians = rois_filtered.median(dim=1).values  # (num_coordinates,)

    # Update the output tensor with medians
    output_batch = batch.clone()
    output_batch[tuple(coords_axes)] = medians
    mask = torch.where(output_batch != batch, 1, 0).to(torch.uint8)

    if struct_params is not None:
        output_batch = _apply_struct_mask_torch(
            output_batch, subpatch_center_coordinates, struct_params
        )

    return output_batch, mask



"""Pydantic CAREamics configuration."""

from __future__ import annotations

import re
from pprint import pformat
from typing import Any, Literal, Union

from bioimageio.spec.generic.v0_3 import CiteEntry
from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator
from typing_extensions import Self

from careamics.config.algorithms import (
    CAREAlgorithm,
    N2NAlgorithm,
    N2VAlgorithm,
)
from careamics.config.data import DataConfig
from careamics.config.training_model import TrainingConfig

ALGORITHMS = Union[
    CAREAlgorithm,
    N2NAlgorithm,
    N2VAlgorithm,
]


class Configuration(BaseModel):
    """
    CAREamics configuration.

    The configuration defines all parameters used to build and train a CAREamics model.
    These parameters are validated to ensure that they are compatible with each other.

    It contains three sub-configurations:

    - AlgorithmModel: configuration for the algorithm training, which includes the
        architecture, loss function, optimizer, and other hyperparameters.
    - DataModel: configuration for the dataloader, which includes the type of data,
        transformations, mean/std and other parameters.
    - TrainingModel: configuration for the training, which includes the number of
        epochs or the callbacks.

    Attributes
    ----------
    experiment_name : str
        Name of the experiment, used when saving logs and checkpoints.
    algorithm : AlgorithmModel
        Algorithm configuration.
    data : DataModel
        Data configuration.
    training : TrainingModel
        Training configuration.

    Methods
    -------
    set_3D(is_3D: bool, axes: str, patch_size: List[int]) -> None
        Switch configuration between 2D and 3D.
    model_dump(
        exclude_defaults: bool = False, exclude_none: bool = True, **kwargs: Dict
        ) -> Dict
        Export configuration to a dictionary.

    Raises
    ------
    ValueError
        Configuration parameter type validation errors.
    ValueError
        If the experiment name contains invalid characters or is empty.
    ValueError
        If the algorithm is 3D but there is not "Z" in the data axes, or 2D algorithm
        with "Z" in data axes.
    ValueError
        Algorithm, data or training validation errors.

    Notes
    -----
    We provide convenience methods to create standards configurations, for instance:
    >>> from careamics.config import create_n2v_configuration
    >>> config = create_n2v_configuration(
    ...     experiment_name="n2v_experiment",
    ...     data_type="array",
    ...     axes="YX",
    ...     patch_size=[64, 64],
    ...     batch_size=32,
    ...     num_epochs=100
    ... )

    The configuration can be exported to a dictionary using the model_dump method:
    >>> config_dict = config.model_dump()

    Configurations can also be exported or imported from yaml files:
    >>> from careamics.config import save_configuration, load_configuration
    >>> path_to_config = save_configuration(config, my_path / "config.yml")
    >>> other_config = load_configuration(path_to_config)

    Examples
    --------
    Minimum example:
    >>> from careamics import Configuration
    >>> config_dict = {
    ...         "experiment_name": "N2V_experiment",
    ...         "algorithm_config": {
    ...             "algorithm": "n2v",
    ...             "loss": "n2v",
    ...             "model": {
    ...                 "architecture": "UNet",
    ...             },
    ...         },
    ...         "training_config": {
    ...             "num_epochs": 200,
    ...         },
    ...         "data_config": {
    ...             "data_type": "tiff",
    ...             "patch_size": [64, 64],
    ...             "axes": "SYX",
    ...         },
    ...     }
    >>> config = Configuration(**config_dict)
    """

    model_config = ConfigDict(
        validate_assignment=True,
        arbitrary_types_allowed=True,
    )

    # version
    version: Literal["0.1.0"] = "0.1.0"
    """CAREamics configuration version."""

    # required parameters
    experiment_name: str
    """Name of the experiment, used to name logs and checkpoints."""

    # Sub-configurations
    algorithm_config: ALGORITHMS = Field(discriminator="algorithm")
    """Algorithm configuration, holding all parameters required to configure the
    model."""

    data_config: DataConfig
    """Data configuration, holding all parameters required to configure the training
    data loader."""

    training_config: TrainingConfig
    """Training configuration, holding all parameters required to configure the
    training process."""

    @field_validator("experiment_name")
    @classmethod
    def no_symbol(cls, name: str) -> str:
        """
        Validate experiment name.

        A valid experiment name is a non-empty string with only contains letters,
        numbers, underscores, dashes and spaces.

        Parameters
        ----------
        name : str
            Name to validate.

        Returns
        -------
        str
            Validated name.

        Raises
        ------
        ValueError
            If the name is empty or contains invalid characters.
        """
        if len(name) == 0 or name.isspace():
            raise ValueError("Experiment name is empty.")

        # Validate using a regex that it contains only letters, numbers, underscores,
        # dashes and spaces
        if not re.match(r"^[a-zA-Z0-9_\- ]*$", name):
            raise ValueError(
                f"Experiment name contains invalid characters (got {name}). "
                f"Only letters, numbers, underscores, dashes and spaces are allowed."
            )

        return name

    @model_validator(mode="after")
    def validate_3D(self: Self) -> Self:
        """
        Change algorithm dimensions to match data.axes.

        Returns
        -------
        Self
            Validated configuration.
        """
        if "Z" in self.data_config.axes and not self.algorithm_config.model.is_3D():
            # change algorithm to 3D
            self.algorithm_config.model.set_3D(True)
        elif "Z" not in self.data_config.axes and self.algorithm_config.model.is_3D():
            # change algorithm to 2D
            self.algorithm_config.model.set_3D(False)

        return self

    def __str__(self) -> str:
        """
        Pretty string reprensenting the configuration.

        Returns
        -------
        str
            Pretty string.
        """
        return pformat(self.model_dump())

    def set_3D(self, is_3D: bool, axes: str, patch_size: list[int]) -> None:
        """
        Set 3D flag and axes.

        Parameters
        ----------
        is_3D : bool
            Whether the algorithm is 3D or not.
        axes : str
            Axes of the data.
        patch_size : list[int]
            Patch size.
        """
        # set the flag and axes (this will not trigger validation at the config level)
        self.algorithm_config.model.set_3D(is_3D)
        self.data_config.set_3D(axes, patch_size)

        # cheap hack: trigger validation
        self.algorithm_config = self.algorithm_config

    def get_algorithm_friendly_name(self) -> str:
        """
        Get the algorithm name.

        Returns
        -------
        str
            Algorithm name.
        """
        return self.algorithm_config.get_algorithm_friendly_name()

    def get_algorithm_description(self) -> str:
        """
        Return a description of the algorithm.

        This method is used to generate the README of the BioImage Model Zoo export.

        Returns
        -------
        str
            Description of the algorithm.
        """
        return self.algorithm_config.get_algorithm_description()

    def get_algorithm_citations(self) -> list[CiteEntry]:
        """
        Return a list of citation entries of the current algorithm.

        This is used to generate the model description for the BioImage Model Zoo.

        Returns
        -------
        List[CiteEntry]
            List of citation entries.
        """
        return self.algorithm_config.get_algorithm_citations()

    def get_algorithm_references(self) -> str:
        """
        Get the algorithm references.

        This is used to generate the README of the BioImage Model Zoo export.

        Returns
        -------
        str
            Algorithm references.
        """
        return self.algorithm_config.get_algorithm_references()

    def get_algorithm_keywords(self) -> list[str]:
        """
        Get algorithm keywords.

        Returns
        -------
        list[str]
            List of keywords.
        """
        return self.algorithm_config.get_algorithm_keywords()

    def model_dump(
        self,
        *,
        mode: Literal["json", "python"] | str = "python",
        include: Any | None = None,
        exclude: Any | None = None,
        context: Any | None = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = True,
        round_trip: bool = False,
        warnings: bool | Literal["none", "warn", "error"] = True,
        serialize_as_any: bool = False,
    ) -> dict:
        """
        Override model_dump method in order to set default values.

        As opposed to the parent model_dump method, this method sets exclude none by
        default.

        Parameters
        ----------
        mode : Literal['json', 'python'] | str, default='python'
            The serialization format.
        include : Any | None, default=None
            Attributes to include.
        exclude : Any | None, default=None
            Attributes to exclude.
        context : Any | None, default=None
            Additional context to pass to the serialization functions.
        by_alias : bool, default=False
            Whether to use attribute aliases.
        exclude_unset : bool, default=False
            Whether to exclude fields that are not set.
        exclude_defaults : bool, default=False
            Whether to exclude fields that have default values.
        exclude_none : bool, default=true
            Whether to exclude fields that have None values.
        round_trip : bool, default=False
            Whether to dump and load the data to ensure that the output is a valid
            representation.
        warnings : bool | Literal['none', 'warn', 'error'], default=True
            Whether to emit warnings.
        serialize_as_any : bool, default=False
            Whether to serialize all types as Any.

        Returns
        -------
        dict
            Dictionary containing the model parameters.
        """
        dictionary = super().model_dump(
            mode=mode,
            include=include,
            exclude=exclude,
            context=context,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
            serialize_as_any=serialize_as_any,
        )

        return dictionary
